import time
import numpy
import tempfile
import copy
import xgboost as xgb
import bytegain.nn_tools.auc_plot as auc_plot
from threading import Thread

def _create_dmatrix(dataset, group_function = None):
    x = dataset._data
    y = dataset._labels
    assert len(x) > 0
    weights = dataset._weights
    matrix = xgb.DMatrix(x, label=y, weight=weights)
    if group_function is not None:
        matrix.set_group(group_function(dataset))
    # Request feature names to fix XGB bug
    names = matrix.feature_names
    return matrix

class XgbClassifier(object):
    def __init__(self, outcomes = 2, threads = 16, loaded_bst = None, group_function = None):
        """

        :param outcomes:
        :param threads:
        :param loaded_bst:
        :param group_function: If set then groups are computed and pairwise ranking is used
        """
        self._outcomes = outcomes
        self._group_function = group_function
        if group_function == None:
            # TODO(jjs) - add support for non-binary outcomes
            self._param = {'silent': True, 'objective':'binary:logistic'}
            self._param['eval_metric'] = 'auc'
        else:
            self._param = {'silent': True, 'objective':'rank:pairwise'}
            self._param['eval_metric'] = 'ndcg'

        if threads is not None:
            self._param['nthread'] = threads
        self._bst = loaded_bst

    def train(self, datasets, max_rounds = 401, min_increase_rounds = 40, l2_norm = 1000, gamma = 1, train_step_size = 0.40, 
        subsample = 0.8, max_depth = 2, min_child_weight = 1, max_delta_step = 0 ):
        dtrain = _create_dmatrix(datasets.train, self._group_function)
        dtest = _create_dmatrix(datasets.test, self._group_function)
        base_rate = numpy.sum(datasets.train.labels) / len(datasets.train.labels)
        print('train base rate: %5.3f%%' % (base_rate * 100))
        evallist  = [(dtest,'eval'), (dtrain,'train')]
        self._param["base_score"] = base_rate
        self._param["early_stopping_rounds"] = min_increase_rounds
        self._param["gamma"] = gamma
        self._param["subsample"] = subsample
        self._param['lambda'] = l2_norm
        self._param['eta'] = train_step_size
        self._param['tree_method'] = 'exact'
        self._param['max_depth'] = max_depth
        self._param['min_child_weight'] = min_child_weight
        self._param['max_delta_step'] = max_delta_step
        self._bst = xgb.train(self._param, dtrain, max_rounds, evallist, xgb_model=self._bst, verbose_eval = 100)

    def get_sorted_results(self, dataset):
        dtest = _create_dmatrix(dataset, self._group_function)
        preds = self._bst.predict(dtest)
        results = []
        if self._group_function is None:
            for entry in zip(preds, dataset._labels, dataset._ids):
                results.append({"probability" : entry[0], "label" : entry[1], "id" : entry[2]})
            auc_plot.sort_results(results)
        else:
            index = 0
            for group_size in self._group_function(dataset):
                end = index + group_size
                group =[]
                for i in range(index, end):
                    group.append({"probability" : preds[i], "label" : dataset._labels[i], "id" : dataset._ids[i]})
                index = end
                auc_plot.sort_results(group)
                results.append(group)

        return results

    def dump_model(self, row_handler, filename):
        fmap = filename + ".fmap"
        create_feature_map(row_handler, fmap)
        self._bst.dump_model(filename, fmap, True)

    def feature_scores(self, row_handler):
        with tempfile.NamedTemporaryFile(delete = False, suffix=".fmap") as fmap:
            create_feature_map(row_handler, fmap.name)
            feature_dict = self._bst.get_score(fmap.name, 'gain')
            sum = 0
            for k,v in list(feature_dict.items()):
                sum += v
            features = [(k, v * 100.0/sum) for k,v in list(feature_dict.items())]
            features = sorted(features, key = lambda x: x[1], reverse = True)
            return features

    # def feature_scores2(self, row_handler):
    #     return xgb_stats.compute_feature_importance(self, row_handler)

    def save(self, filename):
        self._bst.save_model(filename)

def create_feature_map(row_handler, filename):
    with open(filename, "w") as f:
        for i, bucket in enumerate(row_handler._buckets):
            f.write("%d %s q\n " % (i, bucket._name))

def get_params(params):
    k, v = next(iter(list(params.items())))
    new_params = copy.copy(params)
    del new_params[k]
    if type(v) not in (tuple, list):
        v = [v]
    total_runs = []
    if len(new_params) > 0: 
        runs = get_params(new_params)
        for value in v:            
            new_runs = copy.deepcopy(runs)  
            for new_run in new_runs:
                new_run[k] = value
            total_runs.extend(new_runs)
    else:
        for value in v:
            total_runs.append({k : value})
    return total_runs

def run_hyper_param(dataset, params):
    runs = get_params(params)
    print("Total runs: %d" % len(runs))
    outcomes = []
    for run in runs:
        net = XgbClassifier()
        net.train(dataset, **run)
        results = net.get_sorted_results(dataset.test)
        auc = auc_plot.compute_auc(results)
        outcomes.append((auc, run))
        print("AUC: %6.4f %s" % (auc, str(run)))

    outcomes = sorted(outcomes, key = lambda x: x[0], reverse = True)

    for run in outcomes:
        print("AUC: %6.4f %s" % (run[0], str(run[1])))

class CvRunner(Thread):
    def __init__(self, fold, dataset, net, training_cycles, train_args):
        super(CvRunner, self).__init__()
        self._net = net
        self._fold = fold
        self._dataset = dataset
        self._training_cycles = training_cycles
        self._train_args = train_args

    def run(self):
        print("Running CV fold: %d" % self._fold)

        self._net.train(datasets=self._dataset, max_rounds = self._training_cycles, **self._train_args)
        self._sorted_test = self._net.get_sorted_results(self._dataset.test)

class ClonedDatasets(object):
    def __init__(self, datasets):
        self.train = datasets.train.copy()
        self.test = datasets.test.copy()

def cv(dataset, create_net, training_cycles = 361, runs = 5, parallel = False, **train_args):
    start_time = time.time()
    auc_sum = 0
    aucs = []
    test_results = []
    runners = []
    for run in range(runs):
        clone_dataset = ClonedDatasets(dataset) if parallel else dataset
        net = create_net(threads = 3) if parallel else create_net()
        runner = CvRunner(run, clone_dataset, net, training_cycles, train_args)
        runners.append(runner)
        if parallel:
            runner.start()
        else:
            runner.run()

        if run < runs - 1:
            dataset.repartition()

    if parallel:
        for thread in runners:
            thread.join()

    for runner in runners:
        sorted_test = runner._sorted_test
        aucs.append(auc_plot.compute_auc(sorted_test))
        test_results.extend(sorted_test)

    auc_plot.sort_results(test_results)
    auc = auc_plot.compute_auc(test_results)
    print("CV time: %5.3fs" % (time.time() - start_time))
    print("Total Auc: %5.3f %s" % (auc, str([("%5.3f" % a) for a in aucs])))
    return test_results



import pickle
import json
import jsonpickle
import re

# timestamp and anonymous_id are in here because the beam joiner for CSV inputs expects them.
# TODO(chris): The right way to do this would be to pre-join CSV inputs on id.

REQUIRED_COLUMNS = ("id", "received_at", "timestamp", "anonymous_id")

class SchemaInfo(object):
    def __init__(self, schema, main_table, page_table, join_tables):
        self._schema = schema
        self._main_table = main_table
        self._page_table = page_table
        self._join_tables = join_tables

    @property
    def schema(self):
        return self._schema

    @property
    def main_table(self):
        return self._main_table

    @property
    def page_table(self):
        return self._page_table

    @property
    def join_tables(self):
        return self._join_tables

    @property
    def sorted_join_tables(self):
        return sorted(list(self._join_tables.values()), key = lambda v: (v._rows), reverse = True)

    def get_table(self, tablename):
        if tablename == self._main_table.table:
            return self._main_table
        elif tablename == self._page_table.table:
            return self._page_table
        else:
            return self._join_tables[tablename]

    def print_pretty(self):
        print("\nSchema: %s" % self._schema)
        self._main_table.print_pretty()
        self._page_table.print_pretty()

        for join_table in sorted(list(self.join_tables.values()), key = lambda v: v.row_count, reverse=True):
            join_table.print_pretty()

    def remove_table(self, tablename):
        self._join_tables.remove(tablename)

    def remove_columns(self, columns):
        for column_full in columns:
            tablename, columnname = column_full.split(".", 1)
            column_pattern = re.compile(columnname)
            if tablename == "*":
                tables = [self._main_table, self._page_table] + list(self._join_tables.values())
            else:
                tables = [self.get_table(tablename)]
            matches = 0

            for table in tables:
                columns = []
                for column in table.columns:
                    if column_pattern.match(column.name):
                        columns.append(column.name)
                for column in columns:
                    table.remove_column(column)
                matches += len(columns)
            # if matches == 0:
            #       print ("%s not found" % column_full)
            # else:
            #       print ("%s removed %d times" % (column_full, matches))

    def translate_schema_tablenames(self, new_schema):
        """ Translates table names from Blah.XXX to NEW_SCHEMA.Blah_XXX
        """
        self._main_table._table = "%s_%s" % (self._schema, self._main_table._table)
        self._main_table._schema = new_schema
        new_join_table = {}
        for table in list(self._join_tables.values()):
            table._table = "%s_%s" % (self._schema, table._table)
            table._schema = new_schema
            new_join_table[table._table] = table

        self._join_tables = new_join_table
        self._schema = new_schema

class TableInfo(object):
    def __init__(self, schema, table, where = None):
        self._schema = schema
        self._table = table
        self._rows = None
        self._columns = []
        self._column_to_index = None
        self._where = where
        self._s3_manifest = None

    def set_rows(self, rows):
        self._rows = rows

    def set_s3_manifest(self, manifest_url):
        self._s3_manifest = manifest_url

    @property
    def schema(self):
        return self._schema

    @property
    def table(self):
        return self._table

    @property
    def full_name(self):
        return "%s.%s" % (self._schema, self._table)

    @property
    def row_count(self):
        return self._rows

    @property
    def where(self):
        return self._where

    def get_full_where(self, input_where):
        # Hack for backwards compatibility with pickle
        if not hasattr(self, "_where"):
            self._where = None
        input_empty = input_where == None or len(input_where) == 0
        if input_empty and self._where == None:
            return ""
        elif self._where and input_empty:
            return "WHERE %s" % self._where
        elif self._where is None:
            return input_where
        else:
            return "%s AND %s" % (input_where, self._where)

    @property
    def columns(self):
        return self._columns

    def has_column(self, column):
        for x in self.columns:
            if column == x.name:
                return True

        return False

    def get_column(self, columnname):
        index = self.get_column_index(columnname)
        return self._columns[index]

    def get_column_index(self, columnname):

        if not hasattr(self, "_column_to_index") or self._column_to_index == None:
            self._column_to_index = {}         
            for i, column in enumerate(self._columns):
                self._column_to_index[column.name] = i

        return self._column_to_index[columnname]

    def add_column(self, column):
        self._column_to_index = None                
        self._columns.append(column)

    def remove_column(self, columnname):
        column = self.get_column(columnname)
        self._column_to_index = None                
        self._columns.remove(column)

    def __str__(self):
        return "%s: (%s)" % (self.full_name, str(self._rows) if self._rows else "N/A")

    def has_non_required(self):
        for x in self._columns:
            if not x.required():
                return True

        return False

    def print_pretty(self):
        print("\n\n%s" % self)
        if self.row_count > 0:
            for column in self._columns:
                val = column.print_pretty()
                if type(val) != str:
                    print(str(val).encode("utf-8"))


def is_db_type_numeric(datatype):
    return "float" in datatype or "double" in datatype or "int" in datatype or "numeric" in datatype


class ColumnInfo(object):
    def __init__(self, name, datatype):
        self.name = name
        self.datatype = datatype
        self._groups = []
        self.col_length = 0
        self.null_fraction = 0

    def required(self):
        return self.name in REQUIRED_COLUMNS


    def is_numeric(self):
        return is_db_type_numeric(self.datatype)


    def select_stats(self):
        quoted_name = "\"%s\"" % self.name
        selects = ["avg((%s is null)::int::float) as %s_null" % (quoted_name, self.name)]
        if "char" in self.datatype:
            selects.append("avg(length(%s)) as %s_len" % (quoted_name, self.name))
        if self.is_numeric():
            selects.append("avg(%s) as %s_ave, VARIANCE(%s) as %s_var" % (
                quoted_name, self.name, quoted_name, self.name))

        return selects

    def is_groupable(self):
        return ("char" in self.datatype or
                (("int" in self.datatype or "float" in self.datatype) and len(self._groups) > 1 and
                        self.groups_sum() - self._groups[0][1] > 0.005 ))
    def add_group(self, group, fraction):
        self._groups.append((group, fraction))

    def groups_sum(self):
        return sum([x[1] for x in self._groups])

    def read_stats(self, items):
        self.null_fraction = next(items)
        if "char" in self.datatype:
            self.col_length = next(items)
            if self.col_length is None:
                self.col_length = 0
        if self.is_numeric():
            self.average = next(items)
            self.variance = next(items)

    def __str__(self):
        return self.print_pretty()

    def print_pretty(self):
        dt = self.datatype.split(" ")[0]
        if len(dt) > 12:
            dt = dt[0:12]

        if len(self.name) > 30:
            name = self.name[-30:]
        else:
            name = self.name
        base = "  %30s(%s): " % (name, dt)
        if self.null_fraction is None:
            return "%42s: No Info" % (base)

        if "char" in self.datatype:
            base = "%42s %3d%% %3d" % (base, self.null_fraction * 100, self.col_length)
        else:
            base = "%42s %3d%%" % (base, self.null_fraction * 100)

        if self.is_groupable():
            grouping = []
            for group in self._groups:
                fraction = group[1]
                if fraction > 0.001:
                    grouping.append("%s: %2d%%" % (group[0], fraction * 100))

            base = base + " [%s]" % ",".join(grouping)

        return base


def read_old_schema_from_file(filename):
    with open(filename, "rb") as f:
        return pickle.load(f)

def read_schema_from_file(filename):
    with open(filename, "rb") as f:
        schema = jsonpickle.decode(json.dumps(json.load(f)))
        # TODO(jjs) - remove this hack for backwards compatibility
        if not hasattr(schema, '_page_table'):
            schema._page_table = schema.get_table('pages')
            del schema._join_tables['pages']
        return schema

def save_schema(filename, schema, compact = False):
    with open(filename, "wb") as f:
        # indent=0 is just a few percent bigger than indent=None, and make the  half readable.
        # indent=2 is about 2x the size.
        json.dump(json.loads(jsonpickle.encode(schema)), f, sort_keys=True, indent = 0 if compact else 2)




import argparse
import bytegain.custom.data_import.schema_info as schema_info
import bytegain.custom.data_import.db_utils as db_utils
import pickle
import io
import boto3

parser = argparse.ArgumentParser(description='Copies table from redshift to s3')
parser.add_argument('--src_url', type=str, help='db url', required=True)
parser.add_argument('--schema', type = str, help = 'schema of table to copy', required = True)
parser.add_argument('--table', type = str, help = 'table name to copy', required = True)
parser.add_argument('--s3_key_id', type=str, help='AWS user id for S3', required=True)
parser.add_argument('--s3_secret_key', type=str, help='AWS secret key for S3', required=True)
parser.add_argument('--s3_bucket', type=str, help = 'S3 bucket to use', default = "bg-train-input")
parser.add_argument('--dest_folder', type=str, help = 'destination folder', default = "")

parser.add_argument('--where', type=str, help = 'where clause, including "WHERE"', default = "")
parser.add_argument('--orderby', type=str, help = 'order clause including "ORDER BY"', default = "")
parser.add_argument('--timestamp_as_epoch', action="store_true", help = "If set, then dates are saved in Epoch millis",
                    default = False)
parser.add_argument('--no_column_groups', action="store_true",
                    help = "If set, then no computation of group info is done")
parser.add_argument('--columns', default = None)
args = parser.parse_args()


src_conn_factory = db_utils.ConnectionFactory(args.src_url)
cur = src_conn_factory.get_connection().cursor()
table_info = schema_info.TableInfo(args.schema, args.table)
if args.where:
    where = "WHERE %s" % args.where
else:
    where = ""
included_columns = None
if args.columns:
    included_columns = args.columns.split(',')
    columns = [""] * len(included_columns)
else:
    columns = []

table_info = db_utils.compute_table_info(src_conn_factory, table_info, where, no_groups=args.no_column_groups, force_required_columns=False,
                                         included_columns = included_columns)

print("table_info: %s" % table_info)
for column in table_info.columns:
    column_value = '"%s"' % column.name
    if args.timestamp_as_epoch and "timestamp" in column.datatype:
        column_value = "(EXTRACT(EPOCH FROM %s))" % column_value
    elif "char" in column.datatype:
        column_value = "TRANSLATE(%s, '\\r\\n', '  ')" % column_value

    if included_columns:
        if column.name in included_columns:
            columns[included_columns.index(column.name)] = column.name
    else:
        columns.append(column_value)

select = "select %s from %s.%s %s %s" % (", ".join(columns), args.schema, args.table, where, args.orderby)

manifest = db_utils.copy_to_s3_from_redshift_by_bucket(
    cur, select, table_info, args.s3_bucket, args.s3_secret_key, table_info.full_name, args.dest_folder)
local_pickle = io.StringIO()

pickle.dump(table_info, local_pickle)

session = boto3.Session(
    aws_access_key_id=args.s3_key_id,
    aws_secret_access_key=args.s3_secret_key,
)
s3 = session.client('s3')
s3_key = "%s.pickle" % (manifest[0:manifest.find("manifest")])
local_pickle.seek(0)
s3.put_object(Bucket = args.s3_bucket, Key = s3_key, Body = local_pickle)

print("Created manifest: %s" % manifest)



import bytegain.custom.model_data.feature_analysis as feature_analysis
from bytegain.custom.model_data.multi_row_handler import MultiRowHandler
import tensorflow as tf

"""

import bytegain.cust.apartmentlist.al_handler_from_features as al_handler_from_features
import bytegain.model_data.read_from_db as read_from_db
selector = read_from_db.Selector("interest_web_tracks_10", None, "leased_at_notified_property", "interest_id", "original_timestamp")

row_handler = al_handler_from_features.create_multi_row_handler('web_tracks.json')
dataset =       read_from_db.create_datasets(None, 'apartmentlist', selector, row_handler)
"""

# Minimum time for event
MIN_EVENT_TIME = 0.5
INCLUDED_FEATURES = ["context_page_referrer", "event", "current_rental_id", "action", "screen_size", "label", "button_name"]
TIMESTAMP = "original_timestamp"
ID = "id"

# This makes the data files 6x as big, so disabled by default
INCLUDE_ID = False

class ALMultiRowHandler(MultiRowHandler):
    def __init__(self, feature_json, label_bucket = None, id_extractor = None):
        super(ALMultiRowHandler, self).__init__(label_bucket, id_extractor)
        features = feature_analysis.read_features(feature_json)

        for feature in features:
            if feature._column in INCLUDED_FEATURES:
                is_category = feature._datatype._datatype == str
                bucket = feature.create_bucket(is_category = is_category, none_value = -2)
                self.add_bucket(bucket)

    def extra_fields(self):
        if INCLUDE_ID:
            return [TIMESTAMP, ID]
        else:
            return [TIMESTAMP]

    def replace_field_names_with_index(self, cursor):
        super(ALMultiRowHandler, self).replace_field_names_with_index(cursor)
        header = [x.name for x in cursor.description]
        self._time_index = header.index(TIMESTAMP)
        if INCLUDE_ID:
            self._id_index = header.index(ID)

    def _create_event(self, row, elapsed):
        val = []
        for bucket in self._buckets:
            val.extend(bucket.bucketize(row))

        val.append(elapsed)
        return val

    def get_fields(self):
        fields = super(ALMultiRowHandler, self).get_fields()
        fields.append(TIMESTAMP)
        if INCLUDE_ID:
            fields.append(ID)
        return fields

    def _make_data_entry(self):
        last_time = None
        events = []
        for row in self._current_events:
            event_time = row[self._time_index]
            if last_time == None:
                last_time = event_time
                elapsed = 0
            else:
                elapsed = (event_time - last_time).seconds
                last_time = event_time
                if elapsed < MIN_EVENT_TIME:
                    continue

            events.append(self._create_event(row, elapsed))

        return events


    def make_proto_from_row_iterator(self, row_iterator):
        # Filter out later events that fired in short succession.
        def elapsed(first_row, second_row):
            return second_row[self._time_index] - first_row[self._time_index]
        filtered_rows = list(row_iterator)

        context = tf.train.Features(feature={'id': _int64_feature(self.get_id_from_row(filtered_rows[0])),
                                             'label': _int64_feature(int(self.get_label_from_row(filtered_rows[0])[0]))})

        # Add features from bucketes.
        feature_values = {bucket._name: [] for bucket in self._buckets}
        for row in filtered_rows:
            for bucket in self._buckets:
                value = bucket.bucketize(row)
                # TODO: Figure out the right way to do this.  Right now some of these ValueBuckets might actually contain -1s.
                if value == [None]:
                    value = [-1]
                feature = tf.train.Feature(int64_list=tf.train.Int64List(value=value))
                feature_values[bucket._name].append(feature)

        # Add time before next event.
        feature_values['dwell'] = []
        for i in range(len(filtered_rows) - 1):
            feature_values['dwell'].append(tf.train.Feature(int64_list=tf.train.Int64List(value=[elapsed(filtered_rows[i], filtered_rows[i + 1]), 0])))
        feature_values['dwell'].append(tf.train.Feature(int64_list=tf.train.Int64List(value=[0, 1])))

        # Add event id
        if INCLUDE_ID:
            feature_values['id'] = []
            for row in filtered_rows:
                feature_values['id'].append(tf.train.Feature(bytes_list=tf.train.BytesList(value=[row[ID]])))

        feature_lists = tf.train.FeatureLists(feature_list={k: tf.train.FeatureList(feature=v) for k, v in list(feature_values.items())})

        return tf.train.SequenceExample(context=context, feature_lists=feature_lists)

def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))



import bytegain.custom.cust.apartmentlist.al_xgb as al_xgb
import bytegain.custom.nn_tools.xgb_classifier as xgb
import bytegain.nn_tools.auc_plot as auc_plot
import bytegain.custom.cust.apartmentlist.leads_filter_170 as lf
import bytegain.custom.cust.apartmentlist.leads_from_redshift as tfl
feature_file = 'features/al_1.7.0.json'
features = tfl.run_feature_analysis(feature_file)

rh, datasets = al_xgb.create_rh_and_dataset(feature_file, random_seed=2828712, filter = lf.filter2)
classifier = xgb.XgbClassifier()
classifier.train(datasets, max_rounds=401, max_depth=4, l2_norm=400, min_child_weight=100, train_step_size=0.3, subsample=0.9)
features = classifier.feature_scores2(rh)
for feature in features:
    print("%25s: %4.1f" % (feature[0], feature[1]))

before_results = classifier.get_sorted_results(datasets.test)
before_control_results = classifier.get_sorted_results(datasets.get_test_control())

rh, datasets = al_xgb.create_rh_and_dataset(feature_file, random_seed=2828712, filter = lf.filter)
classifier = xgb.XgbClassifier()
classifier.train(datasets, max_rounds=401, max_depth=4, l2_norm=400, min_child_weight=100, train_step_size=0.3, subsample=0.9)
features = classifier.feature_scores2(rh)
for feature in features:
    print("%25s: %4.1f" % (feature[0], feature[1]))

after_results = classifier.get_sorted_results(datasets.test)
after_control_results = classifier.get_sorted_results(datasets.get_test_control())

auc_plot.plot_auc('/tmp/auc_with_prediction_all.png', before_results, results2=after_results,
                  x_label='All Notifications',
                  y_label='Percent of Leases',
                 labels= ["Existing", "With Events"])


auc_plot.plot_auc('/tmp/auc_with_prediction_control.png', before_control_results, results2=after_control_results,
                  x_label='All Notifications',
                  y_label='Percent of Leases',
                  labels= ["Existing", "With Events"])


import argparse
import csv
import json
import logging
import os
import tempfile

import psycopg2
from bytegain.data_import.segment_utils import valid_date
from bytegain.zap.utils import gcs
from google.cloud import storage

logger = logging.getLogger('join_email_from_userid')


def create_connection(dburl=None):
    if dburl is None:
        dburl = os.environ['AL_REDSHIFT_URL']

    return psycopg2.connect(dburl)


def attach_email_to_pred(driver_input, map_config, stages, dataflow_args, input_predictions,
                         output_predictions, tmp_dir):
    """
        Outputs a new CSV which has UserId+email+LTV.
        Input is files with predictions made by cloud ml prediction service containing UserId,LTV
    """
    if not input_predictions.endswith('/'):
        input_predictions += '/'

    if not output_predictions.endswith('/'):
        output_predictions += '/'

    # Download prediction files
    bucket, gs_path = gcs.parse_gs_url(input_predictions)

    client = storage.Client()
    bucket = client.bucket(bucket)
    files = []
    blobs = bucket.list_blobs(prefix=gs_path)
    for blob in blobs:
        filename = blob.name.rsplit('/', 1)[1]
        if filename.startswith('prediction.results'):
            files.append(filename)
    local_files = []
    local_tmp_dir = tempfile.mkdtemp() + '/'
    for filename in files:
        blob = bucket.blob(gs_path + filename)
        local_path = local_tmp_dir + filename
        blob.download_to_filename(local_path)
        local_files.append(local_path)

    # Get users and probabilities
    users = []
    for local_file in local_files:
        with open(local_file, 'r') as input_file:
            for row in input_file:
                row_json = json.loads(row)
                user_id = row_json['id'].rsplit('_', 2)[0]
                prediction = row_json['probability'][0]
                users.append((user_id, prediction))
    users = sorted(users, key=lambda user: user[1], reverse=True)

    logger.info("Read %d users" % len(users))

    assert len(users) > 0

    # Select only users with emails
    start_date = stages['sessions']['data_start_date']

    select = """select user_id, email from api.identifies
    where email is not null and received_at > timestamp '%s'
    order by received_at desc""" % start_date

    emails = {}

    with create_connection() as connection:
        connection.autocommit = False
        with connection.cursor() as cursor:
            cursor.itersize = 100000
            print("Statement: %s" % select)
            cursor.execute(select)
            cursor.fetchone()
            for row in cursor:
                user_id = row[0]
                email = row[1]
                if not user_id in emails:
                    emails[user_id] = email
    logger.info("Read %d users" % len(users))
    output_count = 0

    end_date = stages['sessions']['data_end_date']
    output_filename = 'AL_user_value_predictions_%s-%s_with_emails.csv' % (start_date, end_date)
    output_file = local_tmp_dir + output_filename

    with open(output_file, "w") as o_file:
        writer = csv.writer(o_file)
        for user, ltv in users:
            if user in emails:
                output_count += 1
                writer.writerow([user, emails[user], "%6.4f" % float(ltv)])

    logger.info("Local output file: %s" % output_file)
    logger.info("Cloud output dir: %s" % output_predictions)

    gcs.upload_local_files(local_file_paths=output_file, gs_path=output_predictions)

    gs_output_file = output_predictions + output_filename
    logger.info("Cloud output file: %s" % gs_output_file)

    if output_count < len(users):
        logger.warn("%d of %d users not found" % (len(users) - output_count, len(users)))
    return {'n_users': len(users), 'output_file': gs_output_file}


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('--input', type=str, required=True, help='GCS path to predictions made by CloudML')
    parser.add_argument('--output', type=str, help='Output csv')
    parser.add_argument('--start_date', type=valid_date, help='Start data date')

    args = parser.parse_args()
    stages = {"data_start_date": args.start_date}  # FIXME(yuri): this doesn't look right
    attach_email_to_pred(None, None, stages=stages, dataflow_args=None,
                         input_predictions=args.input, output_predictions='', tmp_dir='/tmp/asa')



from bytegain.zap.prediction_instance import PredictionInstance
import logging
from bytegain.zap.abstract_instance_generator import InstanceGeneratorInterface
import random


def print_session_len(session_duration):
    from bytegain.beam.beam_util import get_counter
    if session_duration < 60:
        get_counter('instance_generator', 'instance_after_mins_00').inc()
    elif session_duration < 3600:
        get_counter('instance_generator',
                    'instance_after_mins_%02d_lease' % (session_duration / 60)).inc()
    elif session_duration < 3600 * 24:
        get_counter('instance_generator', 'instance_after_hrs_%02d' % (
        session_duration / 3600)).inc()
    else:
        get_counter('instance_generator', 'instance_after_days_%02d' % (
        session_duration / 3600 / 24)).inc()


def get_lease_epoch(rows):
    from bytegain.beam.beam_util import get_counter
    from bytegain.zap.abstract_instance_generator import get_timestamp, timestamp_to_epoch

    for row in rows:
        if row.get('type') == 'acappella_leases':
            get_counter('instance_generator', 'acappella_found').inc()
            return timestamp_to_epoch(get_timestamp(row))
    return None


def get_signup_timestamp(rows):
    from bytegain.zap.abstract_instance_generator import get_timestamp

    for row in rows:
        signup = row.get('context', {}).get('page', {}).get('search') == '?signup=true'
        if signup:
            return get_timestamp(row)
    return None


def get_instance_index(rows, instance_end_epoch, include_short_sessions, signup_epoch, lease_epoch, blackout_days):
    from bytegain.beam.beam_util import get_counter
    from bytegain.zap.abstract_instance_generator import get_timestamp, timestamp_to_epoch
    from bytegain.custom.cust.apartmentlist.user_value_model_bg_events.events_to_sessions import quantize

    for row_i in range(len(rows)):
        timestamp_i = timestamp_to_epoch(get_timestamp(rows[row_i]))
        # Decide if make instance even if session ends before `ending_timestamp`
        short_session = (row_i == len(rows) - 1) if include_short_sessions else False
        # Generate instance only if page event
        if rows[row_i].get('type') != 'page':
            continue
        if timestamp_i > instance_end_epoch or short_session:
            # Don't create an instance if lease after 90 days
            if lease_epoch and lease_epoch - timestamp_i > blackout_days * 24 * 3600:
                days_to_lease = (lease_epoch - timestamp_i) / 24 / 3600
                if days_to_lease > 120:
                    get_counter('instance_generator', 'lease_after_120_days').inc()
                else:
                    get_counter('instance_generator', 'lease_after_90_days').inc()
                return None
            else:
                time_since_signup_min = (timestamp_i - signup_epoch) // 60
                # if time_since_signup_min<0:
                #     # No data after sign up
                #     return None
                get_counter('instance_generator', 'time_since_signup: %06d' % quantize(time_since_signup_min)).inc()
                return row_i
    return None


class NewUserValueInstanceGenerator(InstanceGeneratorInterface):
    """
    Predict if a user will sign a lease anywhere (all acappella leases) based on some data after sign up.
    Method specifying how much data after sing up to use `get_instance_end_epoch` should be implemented
    """
    def __init__(self, ending_timestamp_func, include_short_sessions):
        self._blackout_days = 90
        self._ending_timestamp_func = ending_timestamp_func
        # Make instance even if session ends before `ending_timestamp`
        self._include_short_sessions = include_short_sessions

    def get_instances(self, rows):
        """
        :param (list[dict]) rows: a user session
        :return: (list[PredictionInstance]) instances:
        """
        from bytegain.beam.beam_util import get_counter
        from bytegain.zap.abstract_instance_generator import timestamp_to_epoch

        instances = []

        # Find a lease event
        lease_epoch = get_lease_epoch(rows)
        label = True if lease_epoch else False

        # Ignore lease events
        rows = [row for row in rows if row.get('type') != 'acappella_leases']

        # Find if the user signed up
        signup_timestamp = get_signup_timestamp(rows)

        if signup_timestamp is None:
            return instances

        signup_epoch = timestamp_to_epoch(signup_timestamp)

        # Compute instance end time
        instance_end_epoch = self._ending_timestamp_func(signup_timestamp)
        assert instance_end_epoch > signup_epoch

        # Now generate one instance
        # Compute instance index if possible
        instance_index = get_instance_index(rows,
                                            instance_end_epoch=instance_end_epoch,
                                            include_short_sessions=self._include_short_sessions,
                                            signup_epoch=signup_epoch,
                                            lease_epoch=lease_epoch,
                                            blackout_days=self._blackout_days)
        if instance_index:
            instances.append(PredictionInstance(0, instance_index + 1, label))
        else:
            # No data after signup
            get_counter('instance_generator', 'no_data_after_signup').inc()
            # get_counter('instance_generator', 'session_no_instance_length_%04d' % len(rows_no_lease)).inc()
            # logging.info('Session: %s' % rows)
        return instances


def midnight_timestamp(signup_timestamp):
    from bytegain.zap.utils.datetime_utils import from_str
    from datetime import timedelta
    import time

    signup_dt = from_str(signup_timestamp)
    signup_dt_midnight = (signup_dt +
                          timedelta(days=1)).replace(hour=0, minute=0, second=0, microsecond=0)
    return time.mktime(signup_dt_midnight.timetuple())


class NewUserValueMidnightInstanceGenerator(NewUserValueInstanceGenerator):
    """
    Predict if a user will sign a lease (all acappella leases) based on only data on the day when he signed up
    """
    def __init__(self):
        super(NewUserValueMidnightInstanceGenerator, self).__init__(ending_timestamp_func=midnight_timestamp,
                                                                    include_short_sessions=True)


def first_n_minutes_timestamp(signup_timestamp, minutes=4):
    from bytegain.zap.utils.datetime_utils import from_str
    from datetime import timedelta
    import time

    signup_dt = from_str(signup_timestamp)
    signup_dt_midnight = (signup_dt + timedelta(minutes=minutes))
    return time.mktime(signup_dt_midnight.timetuple())


class NewUserValueFirstNMinutesInstanceGenerator(NewUserValueInstanceGenerator):
    """
    Predict if a user will sign a lease (all acappella leases) based on only data on the day when he signed up
    """
    def __init__(self):
        super(NewUserValueFirstNMinutesInstanceGenerator, self).__init__(
            ending_timestamp_func=first_n_minutes_timestamp, include_short_sessions=False)


class NewUserValueFirstNMinutesInstanceGeneratorOld(InstanceGeneratorInterface):
    def __init__(self):
        self._use_minutes = 4
        self._blackout_days = 120

    def get_instances(self, rows):
        """
        Predict if a user will sign a lease anywhere based on only 4 minutes of initial user data

        :param (list[dict]) rows: a user session
        :return: (list[PredictionInstance]) instances:
        """
        from bytegain.beam.beam_util import get_counter
        from bytegain.zap.abstract_instance_generator import timestamp_to_epoch, get_timestamp

        # Find the first lease event
        label = False
        timestamp_lease = None
        for row_i, row in enumerate(rows):
            if row.get('type') == 'acappella_leases':
                get_counter('instance_generator', 'acappella_found').inc()
                label = True
                timestamp_lease = timestamp_to_epoch(get_timestamp(row))
                break

        # Ignore lease events
        rows = [row for row in rows if row.get('type') != 'acappella_leases']

        instances = []
        timestamp_signup = None
        signup = False
        timestamp_prev = timestamp_to_epoch(get_timestamp(rows[0]))
        for row_i in range(1, len(rows)):
            timestamp_i = timestamp_to_epoch(get_timestamp(rows[row_i]))
            if timestamp_signup is not None:
                time_after_signup = timestamp_i - timestamp_signup
                # Create instance after 4 minutes since signing up.
                # NOTE: Use only users who have data after 4 minutes
                if time_after_signup > self._use_minutes * 60:
                    # Don't create an instance if lease after 90 days
                    if timestamp_lease and timestamp_lease - timestamp_i > self._blackout_days * 24 * 3600:
                        days_to_lease = (timestamp_lease - timestamp_i) / 24 / 3600
                        if days_to_lease > 120:
                            get_counter('instance_generator', 'lease_after_120_days').inc()
                        else:
                            get_counter('instance_generator', 'lease_after_90_days').inc()
                        # get_counter('instance_generator', 'lease_after_%03d_days' % days_to_lease).inc()
                    else:
                        time_since_signup_min = (timestamp_prev - timestamp_signup)//60
                        get_counter('instance_generator', 'time_since_signup: %d' % time_since_signup_min).inc()
                        # get_counter('instance_generator', 'time_since_signup_after_4_min: %05d' % ((timestamp_i -
                        #                                                              timestamp_signup)//60)).inc()
                        instances.append(PredictionInstance(0, row_i, label))
                    # One instance per user
                    break
            signup = rows[row_i].get('context', {}).get('page', {}).get('search') == '?signup=true' or signup
            # Use only first signup event
            if signup and timestamp_signup is None:
                timestamp_signup = timestamp_to_epoch(get_timestamp(rows[row_i]))
            timestamp_prev = timestamp_i
        # If haven't created an instance - no data after signup
        if len(instances) == 0:
            get_counter('instance_generator', 'no_data_after_signup').inc()
            rows_no_lease = [row for row in rows if row.get('type')!='acappella_leases']
            # get_counter('instance_generator', 'session_no_instance_length_%04d' % len(rows_no_lease)).inc()
            # logging.info('Session: %s' % rows)
        return instances


class UserValueInstanceGenerator(InstanceGeneratorInterface):
    """Based on all available data predict whether a user will sign a lease in the next X days.
    Make instances every week."""
    def __init__(self):
        lease_window_days = 90
        self._lease_window_secs = lease_window_days * 24 * 3600
        interval_days = 7
        self._interval_secs = interval_days * 24 * 3600
        random.seed(113129321)

    def config(self, driver_input, instance_generator_config, instance_config,
                                  session_generator_data, input_sessions):
        self._start_date = session_generator_data['data_start_date']
        self._end_date = session_generator_data['data_end_date']

    def get_instances(self, rows):
        """
        :param (list[dict]) rows: a user session
        :return: (list[PredictionInstance]) instances:
        """
        from bytegain.beam.beam_util import get_counter
        from bytegain.zap.abstract_instance_generator import timestamp_to_epoch, get_timestamp

        instances = []

        # Find the last non lease event and a lease event (if exists)
        # Ignore sessions which have lease as the first event
        last_before_lease = len(rows) - 1
        timestamp_lease = None
        for row_i, row in enumerate(rows):
            if row.get('type') == 'acappella_leases':
                if row_i == 0:
                    get_counter('instance_generator', 'lease_is_first_event').inc()
                    return instances
                get_counter('instance_generator', 'lease_found').inc()
                timestamp_lease = timestamp_to_epoch(get_timestamp(row))
                last_before_lease = row_i - 1
                break

        # Set the label
        label = False
        if timestamp_lease:
            label = True
            # If there is a big blackout between last non lease event and lease event - this is a negative example
            if timestamp_lease - timestamp_to_epoch(get_timestamp(rows[last_before_lease])) > self._lease_window_secs:
                label = False
                get_counter('instance_generator', 'lease_too_late').inc()

        # # Undersampling
        # if not label and random.randint(0, 10) != 7:
        #     get_counter('instance_generator', 'user_undersampled').inc()
        #     return instances

        # Generate instances every X days. In user is inactive generate 4 instances after he becomes inactive
        row_i = 0
        instance_timestamp = timestamp_to_epoch(get_timestamp(rows[row_i])) + self._interval_secs
        # Last date is either the timestamp of the last event before lease
        # or `end_date` that is specified to get data from BQ
        last_date = get_timestamp(rows[last_before_lease]) if timestamp_lease else self._end_date + 'T00:00:00.000000Z'
        # logging.info("Last date: %s" % last_date)
        last_timestamp = timestamp_to_epoch(last_date)
        # logging.info("Last before lease: %s" % last_before_lease)
        # Number of instances while the user is inactive
        n_instances_no_activity = 0
        while instance_timestamp <= last_timestamp:
            while row_i < last_before_lease and timestamp_to_epoch(get_timestamp(rows[row_i])) < instance_timestamp:
                # logging.info('%s' % get_timestamp(rows[row_i]))
                row_i += 1
                # logging.info("row_i: %s" % row_i)
            final_dwell = instance_timestamp - timestamp_to_epoch(get_timestamp(rows[row_i - 1]))
            instances.append(PredictionInstance(0, row_i, label, final_dwell=final_dwell))
            # logging.info('Instance created with dwell %s' % final_dwell)
            if row_i == last_before_lease:
                n_instances_no_activity += 1
            # Don't generate more than 3 instances when there is no activity
            if n_instances_no_activity > 1 and not label:
                get_counter('instance_generator', 'users_no_activity_3_weeks').inc()
                break
            instance_timestamp += self._interval_secs
        # Add instance last event
        final_dwell = instance_timestamp - timestamp_to_epoch(get_timestamp(rows[last_before_lease]))
        instances.append(PredictionInstance(0, last_before_lease + 1, label, final_dwell=final_dwell))
        # logging.info('Instance created with dwell %s' % final_dwell)
        # Test
        # sess_length_days = (last_timestamp - timestamp_to_epoch(get_timestamp(rows[0]))) / 3600 / 24
        # logging.info("Match: %s" % (sess_length_days//7 == len(instances) - 1))

        get_counter('instance_generator', 'instances_per_user_%02d' % len(instances)).inc()
        return instances




import numpy
import bytegain.custom.model_data.row_handler as row_handler
import xgboost


class ProdXgbClassifier(object):
    def __init__(self, model_spec):
        self._row_handler = row_handler.RowHandler(load_from=model_spec.get_row_handler_file())
        self._bst = xgboost.Booster(model_file = model_spec.get_model_file())
        print("Model %s fields: %s" % (model_spec._model_name, str(self._row_handler.get_fields())))

    def get_fields(self):
        return self._row_handler.get_fields()

    def classify(self, row_dict):
        # Create 1xM input array
        vec = numpy.asarray([self._row_handler.get_data_from_row(row_dict)])
        vec = xgboost.DMatrix(vec)

        predict = self._bst.predict(vec)
        # Return the positive possibility as a native float
        return float(predict[0])

    def classify_batch(self, batch):
        if batch:
            rows = [self._row_handler.get_data_from_row(x) for x in batch]
            vec = numpy.asarray(rows)
            vec = xgboost.DMatrix(vec)
            predicts = self._bst.predict(vec)
            return [float(x) for x in predicts]
        else:
            return [] # prevent blowup on no rows



import argparse
import json
import importlib

from google.cloud import bigquery

import bytegain.custom.cust.apartmentlist.al_xgb as al_xgb
import bytegain.custom.model_data.read_from_db as read_from_db
import bytegain.custom.nn_tools.model_utils as model_utils
from bytegain.custom.model_data.extractor import RowExtractor
from bytegain.custom.nn_tools.prod_classifier import ProdXgbClassifier

parser = argparse.ArgumentParser(description='Tests a model against BQ logs data')
parser.add_argument('--count', type=int, default=1000, help='# examples to verify')
parser.add_argument('--date', type=str, required=True)
parser.add_argument('--model_config', type=str, help = "python var with model config")
parser.add_argument('--source_model', type=str, default = None, help = "Model for source rows, if none use model_config")

args = parser.parse_args()

_bq_client = bigquery.Client(project='bg-rest-api')

def get_config(model_config):
    module, var = model_config.rsplit('.', 1)
    module = importlib.import_module(module)
    return getattr(module, var)

def get_model_name(model_config):
    if args.source_model:
        return args.source_model
    else:
        return "%s_%s" % (model_config['model_name'], model_config['model_version'])

def get_input_rows(config):
    query = """SELECT
raw_score,
features
FROM
`bg-rest-api.al.service_logs`
WHERE
_PARTITIONTIME = TIMESTAMP("{partition_day}")
and model_id='{model_id}'
order by rand() limit {count}
""".format(partition_day = args.date, count=args.count, model_id=get_model_name(config))
    job_config = bigquery.QueryJobConfig()
    # print "BQ query: %s" % query
    query_job = _bq_client.query(query, job_config=job_config)  # API request - starts the query
    iterator = query_job.result()
    rows = list(iterator)
    output_rows = []
    return rows


def load_model(model_name):
    model_spec = model_utils.XgbModelSpec(model_name, path = None)
    model_spec.download(for_prod=False)
    return model_spec, ProdXgbClassifier(model_spec)


def get_local_prediction(classifier, row):
    json_dict = json.loads(row)
    return classifier.classify(json_dict)


def process_lease(input_rows):
    diff = 0
    pred_sum = 0.0
    logs_sum = 0.0
    for row in input_rows:
        prediction = get_local_prediction(classifier, row.features)
        server_prediction = row.raw_score
        pred_sum += prediction
        logs_sum += server_prediction
        delta = server_prediction - prediction
        if abs(delta) > 0.0001:
                    # print ("Delta: %20s %6.4f %6.4f : %7.5f %s" % (id, prediction, train_pred, delta, str(row)))
                    # if classifier is not None:
                    #     print ("input row: %s" % str(classifier._row_handler._row_handler.get_data_from_row(row)))
                    # print "--------------"
                    # misses.append(row)
                    # print ("%5.3f -> %5.3f (%3d%%)" % (server_prediction, prediction, prediction * 100/server_prediction - 100))
                    diff += 1

    count = len(input_rows)
    if count == 0:
        print ("No rows found!")
        exit(1)

    print("%5.3f%% same" % (100 - (diff * 100.0 / count)))
    print("Avg prediction: %6.4f (logs: %6.4f)" % (pred_sum / count, logs_sum / count))
    if count != args.count:
        print("Too few examples: %d" % count)

config = get_config(args.model_config)
model_name = "%s_%s" % (config['model_name'], config['model_version'])

model_spec, classifier = load_model(model_name)
row_handler = classifier._row_handler
input_rows = get_input_rows(config)

print ("Testing against local model")

process_lease(input_rows)




import boto3
import argparse
import bytegain.data_import.segment_utils as segment_utils
import tempfile
import gzip
import shutil
import os
import datetime
import threading
import json

from multiprocessing import Process, Queue
from collections import defaultdict

class WorkerThread(Process):
    def __init__(self, queue, s3_dest, overwrite):
        super(WorkerThread, self).__init__()
        self._queue = queue
        self._s3_dest = s3_dest
        self._overwrite = overwrite
        self._resource = boto3.resource('s3')
        dest_bucket_name, self._dest_key, _ = segment_utils.parse_s3_url(self._s3_dest)
        self._dest_bucket = self._resource.Bucket(dest_bucket_name)

    def run(self):
        ended = False
        while not ended:
            item = self._queue.get()
            if not item:
                ended = True
            else:
                # print ("Got day: %s" % item[1])
                self.create_day(item[0], item[1], item[2])

    def create_day(self, bucket, day, file_list):
        millis = int(day)
        day_timestamp = datetime.datetime.fromtimestamp(millis / 1000.0)

        dest_file = "%s/%s-joined.gz" % (self._dest_key, day_timestamp.strftime("%Y-%m-%d"))
        if not self._overwrite:
            response = self._dest_bucket.objects.filter(Prefix=dest_file)
            count = 0
            for item in response:
                # print "item: %s" % str(item)
                count += 1
            if count > 0:
                print("Skipping %s" % dest_file)
                return

        my_bucket = self._resource.Bucket(bucket)
        tfiles = []
        for key in file_list:
            tfile = tempfile.NamedTemporaryFile(delete = False)
            tfile.close()
            my_bucket.download_file(key, tfile.name)
            tfiles.append(tfile.name)

        output_file = tempfile.NamedTemporaryFile(delete = False, suffix = ".gz")
        output_file.close()
        with gzip.open(output_file.name, 'wb') as output_gzip:
            for file in tfiles:
                with gzip.open(file, 'rb') as f:
                    shutil.copyfileobj(f, output_gzip)
                    os.remove(file)

        print("Uploading %s to %s" % (output_file.name, dest_file))
        self._dest_bucket.upload_file(output_file.name, dest_file)
        os.remove(output_file.name)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Copies and Joins S3 data files to Bytegain S3 storage')
    parser.add_argument('--s3_input', type=str, help='S3 input directory', required=True)
    parser.add_argument('--s3_output', type=str, help='S3 output directory', required=True)
    parser.add_argument('--s3_profile', type=str, help='s3 profile to use for data access', default = "s3_data_user")
    parser.add_argument('--threads', type=int, help='number of parallel S3 threads', default = 4)
    parser.add_argument('--overwrite', help='If set, then overwrite destination files, otherwise skip existing days', action='store_true')

    args = parser.parse_args()
    boto3.setup_default_session(profile_name=args.s3_profile)
    session = boto3.Session()
    s3 = session.client('s3')

    bucket, key, _ = segment_utils.parse_s3_url(args.s3_input)
    start_key = ""
    day_files = defaultdict(lambda: [])
    files = 0
    while (True):
        response = s3.list_objects_v2(Bucket=bucket, Prefix=key, StartAfter = start_key)
        if "Contents" not in response:
            break

        contents = response["Contents"]
        for file in contents:
            start_key = file["Key"]
            if ".gz" in start_key:
                files += 1
                dirs = start_key.split('/')
                day = dirs[-2]
                day_list = day_files[day]
                day_list.append(start_key)

    print("%d files in %d days" % (files, len(day_files)))

    file_queue = Queue(20)
    processes = []
    for i in range(args.threads):
        thread = WorkerThread(file_queue, args.s3_output, args.overwrite)
        thread.start()
        processes.append(thread)

    for day in sorted(day_files.keys()):
        file_queue.put([bucket, day, day_files[day]])

    for process in processes:
        process.join()



class FeatureBucketerLinear(object):
    """
    feature_list: TableInfo object
    excluded_features: list
    """
    def __init__(self, excluded_features, feature_list):
        self._excluded_features = set(excluded_features)
        self._features = feature_list
        self._buckets = {}
        self._counts = {'categorical': 0, 'numerical': 0, 'boolean': 0, 'time': 0, 'excluded': 0}
        self._feature_map = {}
        self._total_length = 0
        self._num_buckets = [17, 20] # [Numerical buckets without inf and None, Total number of numerical buckets]

    @property
    def total_length(self):
        return self._total_length

    @property
    def feature_map(self):
        return self._feature_map

    def construct_buckets(self):
        """
        Construct buckets for all features based on the schema from TableInfo file

        Bucketing logic: Each feature is encoded into a one-hot vector. All one-hots are concatenated together into a
        large sparse vector

        Assumption: there are only 4 data types: character - categorical, boolean, time-like and numerical

        How does it work?
        1) "construct_buckets" stage:
        - For each feature we construct two maps: one-hot encoding map and feature index map

        One hot encoding map:
         - e.g. boolean features are associated with the following map
         `bucketer._buckets['boolean_feature'] = {'f': 0, 't': 1, '': 2, None: 2}`
         - Map's width is determined either by the nature of the feature (e.g. boolean features always
         have width 3) or is set by a user (e.g. for numerical features we choose the bucket width)

        Feature index map:
         - Each feature gets its position in the final output vector, "feature index"
            {feature1, feature2, ..., featureN} ->  {index1, index2, ..., indexN}

        2) "sample_to_vec" stage:
         - One-hot and feature maps are combined to form one large sparse vector, which encodes all features
         {feature_1, feature_2, ..., feature_N} - > [one_hot_1, one_hot_2, ..., one_hot_N] = output
        :return:
        """

        bucket_index = 0
        for col in self._features.columns:
            print("Feature: %s" % col)
            if col.name in self._excluded_features:
                self._counts['excluded'] += 1
                continue
            # Categorical features
            if 'character' in col.datatype:
                buckets = self.category_buckets(sorted(col._groups, key=lambda x: x[1]))
                self._buckets[col.name] = buckets
                bucket_len = len(buckets)
                self._counts['categorical'] += 1
            # Boolean features
            elif col.datatype == 'boolean':
                # Treat None or '' as a separate state
                self._buckets[col.name] = {'f': 0, 't': 1, '': 2, None: 2}
                bucket_len = 4
                self._counts['boolean'] += 1
            # Time-like features - ignore them
            elif 'time' in col.datatype:
                # Add time-like features into excluded set
                self._excluded_features.add(col.name)
                bucket_len = 0
                self._counts['time'] += 1
                continue
            # Numerical features
            else:
                stats = col.median, col.percentile_25, col.percentile_75
                buckets = self.numerical_buckets(stats)
                self._buckets[col.name] = buckets
                bucket_len = self._num_buckets[1]
                self._counts['numerical'] += 1
            self._feature_map[col.name] = bucket_index
            bucket_index += bucket_len
        self._total_length = bucket_index
        print("Bucketed features: %d, total number of features: %d" % (len(self._buckets), len(self._features.columns)))
        print("Length of buckets: %d" % self._total_length)
        # Print features used
        print("Features used")
        print("=========================================================")
        for k, v in list(self._buckets.items()):
            print("Feature: %s, value: %s" % (k, v))
        # return len(self._buckets)

    def category_buckets(self, groups):
        """
        Buckets for categorical features
        :param groups:
        :return:
        """
        # Category -> index map
        cat_ind_map = {}
        for index, (name, fraction) in enumerate(groups):
            # Map `None` to ''
            if not name:
                name = ''
            cat_ind_map[name] = index
        # Treat None or '' as a separate state
        if '' not in cat_ind_map:
            index += 1
            cat_ind_map[''] = index
        index += 1
        cat_ind_map['other'] = index
        return cat_ind_map

    def numerical_buckets(self, stats):
        """
        Buckets for numerical features
        :param stats:
        :param num_buckets:
        :return:
        """
        import numpy as np
        (median, perc25, perc75) = stats
        bins = np.linspace(perc25, perc75, self._num_buckets[0])
        bins = np.insert(bins, 0, -np.inf)
        bins = np.append(bins, np.inf)
        return bins

    def sample_to_vec(self, sample):
        """
        Encode features into one large sparse vector
        :param sample:
        :return: sparse vector
        """
        import logging
        import numpy as np
        sample_vec = [0]*self._total_length
        bucketed_features = set([])
        all_features = set([])
        total_length = 0
        for k, v in list(sample.items()):
            all_features.add(k)
            if k in self._excluded_features:
                continue
            # print 'Feature: %s' % k
            bucket_map = self._buckets.get(k, [])
            if len(bucket_map)<1:
                continue
            index0 = self._feature_map[k]
            # Categorical and bool features
            if isinstance(bucket_map, dict):
                v = '' if v is None else v
                index = index0 + bucket_map.get(v)
                bucket_len = len(bucket_map)
                if index is None:
                    logging.info("Index None for feature :%s, value %s, bucket map %s" % (k, v, bucket_map))
            else:
                # print 'Numerical bucket length %d' % len(bucket_map)
                # Treat None as a separate state
                bucket_len = len(bucket_map) + 1
                if v == '':
                    index = index0 + bucket_len - 1
                # Other numerical
                else:
                    index = index0 + np.digitize(float(v), bucket_map) - 1
            total_length += bucket_len
            # print 'Feature %s, bucket length %d' % (k, bucket_len)
            if bucket_len > 0:
                sample_vec[index] = 1
                bucketed_features.add(k)
        # print 'Total Length %d' % total_length
        return sample_vec


class FeatureBucketerNonLinear(FeatureBucketerLinear):
    """
    Inherits from FeatureBucketerLinear and overrides numerical feature binning
    """
    def __init__(self, excluded_features, feature_list, n_numerical_buckets=10):
        super(FeatureBucketerNonLinear, self).__init__(excluded_features, feature_list)
        self._num_buckets = [n_numerical_buckets - 1, n_numerical_buckets]  # Reserve one bucket for None

    def numerical_buckets(self, stats):
        """
        Buckets for numerical features
        :param stats: statistics extracted during data generation
        :param num_buckets:
        :return:
        """
        beta, alpha = 0.8, 0.2
        import numpy as np
        (median, perc25, perc75) = stats
        b = np.log((1/alpha - 1)/(1/beta - 1)) / (perc75 - perc25)
        a = np.exp(b * perc25) * (1 / alpha - 1)
        bins = np.linspace(0, 1, self._num_buckets[0])
        # print 'a: %s, b: %s' % (a, b)
        return a, b, bins

    def sample_to_vec(self, sample):
        """
        Encode features into one large sparse vector
        :param sample:
        :return:
        """
        import logging
        import numpy as np
        # Sparse output feature vector
        sample_vec = [0] * self._total_length
        # For stats
        bucketed_features = set([])
        all_features = set([])
        total_length = 0
        # Set a bit for each feature
        for k, v in list(sample.items()):
            all_features.add(k)
            if k in self._excluded_features:
                continue
            # print 'Feature: %s' % k
            # Get on-hot map
            bucket_map = self._buckets.get(k, [])
            if len(bucket_map) < 1:
                continue
            # Get feature index
            index_feature = self._feature_map[k]
            if index_feature is None:
                logging.error("Feature bucket index is None for feature: %s, value: %s, "
                              "bucket map: %s" % (k, v, bucket_map))
            # Categorical and boolean features
            if isinstance(bucket_map, dict):
                v = '' if v is None else v
                # Compute index within the bucket
                index_cur = bucket_map.get(v) if v in bucket_map else bucket_map.get('other')
                if index_cur is None:
                    logging.error("Feature index within bucket is None for feature: %s, value: %s, "
                                  "bucket map: %s, sample: %s" % (k, v, bucket_map, sample))
                index = index_feature + index_cur
                bucket_len = len(bucket_map)
            # Numerical features
            else:
                bucket_len = len(bucket_map[2]) + 1  # bucket_map returns `(a, b, bins)`, we need only bins
                # Treat None as a separate state
                if v == '':
                    index = index_feature + bucket_len - 1
                # Other numerical values
                else:
                    a, b, bins = bucket_map
                    func = 1 / (1 + a * np.exp(- b * float(v)))
                    index = index_feature + np.digitize(func, bins) - 1
            total_length += bucket_len
            # print 'Feature %s, bucket length %d' % (k, bucket_len)
            if bucket_len > 0:
                sample_vec[index] = 1
                bucketed_features.add(k)
        # print 'Total Length %d' % total_length
        return sample_vec


class FeatureBucketerDirect(FeatureBucketerLinear):
    def __init__(self, excluded_features, feature_list):
        super(FeatureBucketerDirect, self).__init__(excluded_features, feature_list)

    # Construct buckets for all features based on TableInfo
    def construct_buckets(self):
        # Assume that there are only 4 data types: character - categorical, boolean, time-like and numerical
        total_length = 0
        for col in self._features.columns:
            if col.name in self._excluded_features:
                self._counts['excluded'] += 1
                continue
            print('Feature: %s' %col.name)
            if 'character' in col.datatype:
                buckets = self.category_buckets(sorted(col._groups, key=lambda x: x[1]))
                self._buckets[col.name] = buckets
                bucket_len = len(buckets)
                self._counts['categorical'] += 1
            elif col.datatype == 'boolean':
                # Treat None or '' as a separate state
                self._buckets[col.name] = {'f': 0, 't': 1, '': 2, None: 2}
                bucket_len = 4
                self._counts['boolean'] += 1
            elif 'time' in col.datatype:
                self._counts['time'] += 1
                # Add time-like features into excluded set
                self._excluded_features.add(col.name)
                bucket_len = 0
            else:
                self._counts['numerical'] += 1
                self._buckets[col.name] = [0, 0]
                bucket_len = 2
            total_length += bucket_len
            # print 'Feature %s, bucket length %d' % (col.name, bucket_len)
        print(("Bucketed features: %d, total # of features: %d, excluded: %d " % (
        len(self._buckets), len(self._features.columns), len(self._excluded_features))))
        print("# of buckets: %d" % total_length)
        # Print features used
        print("Features used")
        print("=========================================================")
        for k, v in list(self._buckets.items()):
            print("Feature: %s, value: %s" % (k, v))
            # return len(self._buckets)

    # Bucket features
    def sample_to_vec(self, sample):
        import logging
        import numpy as np
        sample_vec = []
        bucketed_features = set([])
        all_features = set([])
        total_length = 0
        for k, v in list(sample.items()):
            all_features.add(k)
            if k in self._excluded_features:
                continue
            # print 'Feature: %s' % k
            bucket_map = self._buckets.get(k, [])
            if len(bucket_map) < 1:
                continue
            # Categorical and bool features
            if isinstance(bucket_map, dict):
                v = '' if v is None else v
                index = bucket_map.get(v)
                bucket_len = len(bucket_map)
                if index is None:
                    logging.info("Index None for feature :%s, value %s, bucket map %s" % (k, v, bucket_map))
                vec = [0] * bucket_len
                vec[index] = 1
            else:
                bucket_len = 2
                vec = [0] * bucket_len
                if v == '':
                    vec[0] = 1
                # Other numerical
                else:
                    vec[1] = float(v)
            bucketed_features.add(k)
            sample_vec += vec
            total_length += bucket_len
        return sample_vec



import argparse
import os.path

from bytegain.custom.model_data import feature_analysis

THRESHOLD = 0.05
THRESHOLD_MSG = '> %.0f%%' % (THRESHOLD * 100)
CAT_THRESHOLD = 0.1

IGNORED_FEATURES = {}


def round_if_float(v):
    if type(v) == float:
        return round(v, 4)
    return v


def approx_equal(val1, val2):
    if (val1 < 0 < val2) or (val2 < 0 < val1):
        return False
    val1 = abs(val1)
    val2 = abs(val2)
    p = abs(val1 - val2)
    if val1 > 1 or val2 > 1:
        p = p / float(max(val1, val2))
    return p <= THRESHOLD


def approx_cat_match(cat1, cat2):
    if len(cat1) != len(cat2):
        return False
    for idx, c1 in enumerate(cat1):
        c2 = cat2[idx]
        if round_if_float(c1[0]) != round_if_float(c2[0]) or abs(c1[1] - c2[1]) > CAT_THRESHOLD:
            return False
    return True


def compare_features(current_features, current_label, previous_features, previous_label):
    print('>>> Feature diff:')
    print('1st line: %s' % current_label)
    print('2nd line: %s' % previous_label)

    broken_features = []

    def add_broken_feature(feature, reason):
        broken_features.append(feature)
        print('^^^ DIFF: %s' % reason)

    previous_features_dict = {}
    for feature_info in previous_features:
        previous_features_dict[feature_info.name()] = feature_info
    for current_fi in current_features:
        previous_fi = previous_features_dict.get(current_fi.name())
        print('')
        print(current_fi)
        print(previous_fi)

        if previous_fi is None:
            add_broken_feature(current_fi.name(), 'previous data does not exist')
        elif current_fi.data_type() != previous_fi.data_type():
            add_broken_feature(current_fi.name(), 'different data types')
        elif not approx_equal(current_fi.null_fraction(), previous_fi.null_fraction()):
            add_broken_feature(current_fi.name(), 'null fraction %s' % THRESHOLD_MSG)
        elif current_fi.name() in IGNORED_FEATURES:
            continue
        elif current_fi.data_type().is_cat():
            # categorical
            previous_top_vals_dict = {}
            for val in previous_fi.top_values():
                v = round_if_float(val[0])
                previous_top_vals_dict[v] = val[1]
            for val in current_fi.top_values():
                cur_val = round_if_float(val[0])
                cur_fraction = val[1]
                prev_fraction = previous_top_vals_dict.get(cur_val, 0.0)
                if not approx_equal(cur_fraction, prev_fraction):
                    if not approx_cat_match(current_fi.top_values(), previous_fi.top_values()):
                        add_broken_feature(current_fi.name(), 'value: %s' % cur_val)
                    break
        elif current_fi.data_type().is_number():
            # non-categorical number
            if (not approx_equal(current_fi.average(), previous_fi.average()) and
                    not approx_equal(current_fi.median(), previous_fi.median())):
                add_broken_feature(current_fi.name(), 'avg and median %s' % THRESHOLD_MSG)
        else:
            # non-categorical something else -- anything to check?
            pass

    if broken_features:
        print('\n>>> Features requiring attention:\n%s' % '\n'.join(broken_features))

    return broken_features


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Compare feature analysis stored as local JSON files')
    parser.add_argument('request_features_file', help='Local file with request feature analysis')
    parser.add_argument('training_features_file', help='Local file with training feature analysis')
    args = parser.parse_args()
    if compare_features(feature_analysis.read_features(args.request_features_file),
                        'request logs (%s)' % os.path.basename(args.request_features_file),
                        feature_analysis.read_features(args.training_features_file),
                        'training (%s)' % os.path.basename(args.training_features_file)):
        exit(1)




class FeatureFilter(object):
    def __init__(self, excluded, numerical, categorical, bucket = None):
        self._excluded = set(excluded)
        self._numerical = set(numerical)
        self._categorical = set(categorical)
        self._bucket = set(bucket)

    def create_buckets(self, features):
        buckets = []
        output_features = []
        for feature in features:
            if self.excluded(feature):
                print("Explicitly Excluding %s" % str(feature))
            elif feature.is_suitable() or self.categorical(feature):
                is_category = True if self.categorical(feature) else (False if self.numerical(feature) else None)
                is_one_hot = self.bucket(feature)
                bucket = feature.create_bucket(is_category = is_category, none_value = -2, one_hot = is_one_hot)
                buckets.append(bucket)
                output_features.append((feature, bucket))
            else:
                print("Not suitable: %s %s" % (str(feature), feature._column))

        print ("\n--------- Features -------------")
        for feature, bucket in output_features:
            print("%s  --> %s" % (feature, type(bucket)))

        return buckets

    def excluded(self, feature):
        return feature._column in self._excluded

    def numerical(self, feature):
        return feature._column in self._numerical

    def categorical(self, feature):
        return feature._column in self._categorical

    def bucket(self, feature):
        return feature._column in self._bucket



import argparse
import os
import datetime
from bytegain.nn_tools.verify_past_predictions import  verify_predictions
import bytegain.custom.data_import.db_utils as db_utils
import bytegain.data_import.segment_utils as segment_utils


def get_outcomes(start_date, predictions):

    src_conn_factory = db_utils.ConnectionFactory(os.environ['AL_REDSHIFT_URL'])
    cur = src_conn_factory.get_connection().cursor()

    select = """
select user_id from etl.acappella_leases 
where created_at > %(date)s order by 1"""
    cur.execute(select, {'date': start_date})

    outcomes = {}
    for k in list(predictions.keys()):
        outcomes[k] = 0
    positives = 0
    for row in cur:
        user_id = str(row[0])
        if user_id in predictions:
            outcomes[user_id] = 1
            positives += 1

    return outcomes, positives


if __name__ == '__main__':
    parser = argparse.ArgumentParser("Takes predictions made in the past and verifies them using new data.")
    parser.add_argument('--prediction_made_date', type = segment_utils.as_date,
                        required=True, help='Date when predictions_where made')
    args = parser.parse_args()
    previous_week = args.prediction_made_date -datetime.timedelta(days=7)
    path = "gs://bg-dataflow/apartmentlist/user_value/predictions/{start_date}-{end_date}/AL_user_value_predictions_{start_date}-{end_date}.json"
    path = path.format(start_date=previous_week.date().isoformat(), end_date=args.prediction_made_date.date().isoformat())
    verify_predictions(path, lambda predictions: get_outcomes(args.prediction_made_date, predictions))






import bytegain.custom.model_data.read_from_db as read_from_db
import bytegain.custom.model_data.feature_analysis as feature_analysis
import bytegain.custom.cust.apartmentlist.al_handler_from_features as al_handler_from_features
import csv

def get_selector():
    return read_from_db.Selector("interest_web_tracks_10", None, "leased_at_notified_property", "interest_id", "original_timestamp")

def create_csv_from_row_handler(filename, password = None, row_limit = None):
    selector = get_selector()
    row_handler = al_handler_from_features.create_multi_row_handler('web_tracks.json')
    read_from_db.generate_row_handler_stats(password, 'apartmentlist', selector, row_handler)
    data,labels,ids,rh  = read_from_db.create_datasets(password, 'apartmentlist', selector, row_handler, row_limit = row_limit)

    with open(filename, "w") as f:
        writer = csv.writer(f, delimiter='|')
        for event,label,id in zip(data, labels, ids):
            row = [id, int(label[0])]
            row.extend([",".join([str(e) for e in x]) for x in event])
            writer.writerow(row)

def run_feature_analysis(filename):
    data_info = feature_analysis.DataInfo('apartmentlist', get_selector(), None)
    features = feature_analysis.process_features(data_info)
    feature_analysis.save_features(filename, features)
    return features



EXCLUDED_FEATURES = [
    # 'search_category',
    # 'viewed_availability',
    #'clicked_email',
    'opened_email',
    # 'preferences_lease_length',
    # 'preferences_move_urgency',

    'email_opens_count',
    'email_clicks_count',
    # 'median_matching_price',
    'matching_units',
    'all_price',
    'availability_views',
    'commute_congestion',
    'preferences_amenities',
    'preferences_beds',
    'score',
    'interest_id',
    'leased_at_notified_property',
    'matched_user',
    'commute_lat',
    'commute_lon',
    'source_app',
    # median_matching_price replaced by median_preferred_price
    'median_matching_price',
    'photo_gallery_views',
    # Prev interests excluded because it's hard to compute at runtime
    'prev_interests',
    'consent_call_consent',
    'consent_email_consent',
    'preferences_contracts',
    'renter_phone_area_code',
    'renter_phone_exchange',
    'interest_created_at',
    'total_units',
    'matched_interest',
    'user_id',

    'preferences_wants_dishwasher',
    'preferences_most_important_feature',
    'preferences_wants_on_site_laundry',
    'preferences_wants_gym',
    'preferences_wants_in_unit_laundry',
    'preferences_wants_dog_friendly',
    'preferences_wants_pool',
    'preferences_wants_air_conditioning',
    'preferences_wants_parking',
    'preferences_cotenant',
    'preferences_cosigner',
    'avg_12mo_2bd_location_price',
    'relative_2bd_price',

    # 'preferences_baths',
    # 'preferences_beds_0',
    # 'preferences_beds_1',
    #'preferences_beds_2',
    # 'preferences_beds_3',
    # 'has_phone_number',
    # 'email_clicked',
    # 'commute_mode',
    # 'registered_source_app',
    'property_lat_round',
    'property_lon_round',
    # 'relative_positive_velocity_rate',
    'property_to_total_velocity_rate',
    'positive_interest_velocity',
    'total_interest_velocity',
    # 'prediction',
    'score',
    'model',
    'version',
    'rental_id',
    'label',
    'extra_label',
    'leased_at_property',
    'value_on',
    'event_id',
    'property_city_id',
    'email_domain',
    'registered_source_partner',
    'registered_source_app',
    'interest_source_app',
    'property_zip_code',
    'metro_id',
    'property_lon',
    'property_lat'
]


import bytegain.custom.model_data.feature_analysis as feature_analysis


finance_features = feature_analysis.read_features('al_finance_v1.0.0')
interest_features = feature_analysis.read_features('interest_created_1h_v1.1.0')

interest_feature_dict ={}
for feature in interest_features:
    interest_feature_dict[feature._column] = feature

for feature in finance_features:
    if feature._column not in interest_feature_dict:
        print("Missing feature: %s" % feature._column)
        continue

    interest_feature = interest_feature_dict[feature._column]
    print("\n\nFeature: %s"  % feature._column)
    for i in range(len(feature.top_values())):
        if i >= len(interest_feature.top_values()):
            break
        print("%s %s" % (feature.top_values()[i], interest_feature.top_values()[i]))



import argparse
import csv


def read_predictions(file):
    """
    Read csv with user_id, email and prediction into a hash table
    :param file: (string) Path to a local file.
    :return: predictions, max_val: (dict, float) Return hash table with predictions and rankings along with
    maximum prediction
    """
    predictions = {}
    max_val = 0.0
    with open(file, mode='r') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        for row_i, row in enumerate(csv_reader):
            user_id = row[0]
            pred = float(row[2])
            predictions[user_id] = {'pred': pred, 'rank': row_i}
            if pred > max_val:
                max_val = pred
    return predictions, max_val


def compare_predictions(file1, file2):
    """
    Compare predictions from two csv files
    :param file1: (string) Path to file1
    :param file2: (string) Path to file2
    :return:
    """
    results1, max_val1 = read_predictions(file1)
    results2, max_val2 = read_predictions(file2)

    average_norm_diff = 0.0
    average_diff = 0.0
    average_rank_diff = 0.0
    for key in list(results1.keys()):
        pred1 = results1[key]['pred']
        pred2 = results2[key]['pred']

        rank1 = results1[key]['rank']
        rank2 = results2[key]['rank']

        average_norm_diff += abs(pred1/max_val1 - pred2/max_val2)
        average_diff += abs(pred1 - pred2)
        average_rank_diff += abs(rank1 - rank2)
    print("Average difference in normalized predictions: %.4f" % (average_norm_diff / len(results1)))
    print("Average difference in absolute predictions: %.4f" % (average_diff / len(results1)))
    print("Average difference in ranking: %.4f" % (average_rank_diff / len(results1)))


if __name__ == '__main__':
    parser = argparse.ArgumentParser('')
    parser.add_argument('--file1', required=True)
    parser.add_argument('--file2', required=True)

    args = parser.parse_args()

    compare_predictions(args.file1, args.file2)


# This compiles & loads Cython (pyx) modules if the code is more recent than .so
import pyximport

pyximport.install()

import itertools
import numpy
import math
import os.path
import bytegain.custom.model_data.feature_analysis as feature_analysis
import bytegain.custom.model_data.read_from_db as read_from_db
import bytegain.custom.nn_tools.xgb_classifier as xgb
import bytegain.custom.nn_tools.model_utils as model_utils
from bytegain.custom.model_data.row_handler import RowHandler
from bytegain.custom.model_data.bucket import ValueBucket
from bytegain.custom.model_data.extractor import RowExtractor


class SearchIdExtractor(object):
    def __init__(self):
        self._user_id_index = None
        self._user_order_index = None
        self._search_category_index = None

    def extract(self, row):
        if self._user_id_index is not None:
            search_category = row[self._search_category_index]
            if search_category is None or "_" in search_category:
                search_category = "NA"
            return "%s_%s_%s" % (row[self._user_id_index], search_category, row[self._user_order_index])
        else:
            return "%s_%s_%s" % (row['user_id'], row['search_category'], row['user_order'])

    def get_fields(self):
        return ['user_id', 'user_order', 'search_category']

    def update_fields_by_id(self, cursor_header):
        self._user_id_index = cursor_header.index('user_id')
        self._user_order_index = cursor_header.index('user_order')
        self._search_category_index = cursor_header.index('search_category')


class InterestStateExtractor(RowExtractor):
    def __init__(self):
        super(InterestStateExtractor, self).__init__("final_interest_state")

    def update_fields_by_id(self, cursor_header):
        super(InterestStateExtractor, self).update_fields_by_id(cursor_header)

    def extract(self, row):
        state = super(InterestStateExtractor, self).extract(row)
        return state in ('visit_interest', 'strong_interest', 'mild_interest')

    def get_fields(self):
        return ['final_interest_state']


class NotificationAndInterestStateExtractor(RowExtractor):
    def __init__(self):
        super(NotificationAndInterestStateExtractor, self).__init__("final_interest_state")
        self._leased_extractor = RowExtractor("notified_at")

    def update_fields_by_id(self, cursor_header):
        super(NotificationAndInterestStateExtractor, self).update_fields_by_id(cursor_header)
        self._leased_extractor.update_fields_by_id(cursor_header)

    def extract(self, row):
        state = super(NotificationAndInterestStateExtractor, self).extract(row)
        notified_at = self._leased_extractor.extract(row)
        if notified_at:
            return 3
        if state in ('visit_interest', 'strong_interest'):
            return 2
        elif state == 'mild_interest':
            return 1
        else:
            return 0

    def get_fields(self):
        return ['final_interest_state', "notified_at"]


def run_feature_analysis(filename):
    data_info = feature_analysis.DataInfo(get_selector())
    features = feature_analysis.process_features(data_info)
    feature_analysis.save_features(filename, features)
    return features


def get_dataset(feature_file, filter, label_extractor, run_analysis=False, random_seed=358377213, rows = None):
    if run_analysis:
        features = run_feature_analysis(feature_file)
        print("Features: %s" % features)
    else:
        features = feature_analysis.read_features(feature_file)
    rh = create_row_handler(features, filter, label_extractor)
    datasets = create_datasets_from_row_handler(rh, test_fraction=0.2, random_seed=random_seed, rows=rows)
    return rh, datasets


def get_selector():
    return read_from_db.Selector("apartmentlist.search_training",
                                 # None,
                                 "interest_created_at > timestamp '2017-08-15'",
                                 # "final_interest_state = 'strong_interest'",
                                 "(final_interest_state in  ('visit_interest', 'strong_interest', 'mild_interest'))",
                                 # "(final_interest_state in  ('visit_interest', 'strong_interest'))",
                                 "md5(user_id)")

def create_datasets_from_row_handler(row_handler, random_seed=83243312, test_fraction=0.2, rows=None):
    selector = get_selector()
    numpy.random.seed(random_seed)
    read_from_db.generate_row_handler_stats(selector, row_handler)
    return read_from_db.create_datasets(selector, row_handler, test_fraction, rows, shuffle_rows = False)


def compute_NGDC(results, for_model = True, by_category = False, print_users = 0):
    # First sort by ranking_key, then sort by user_id

    if for_model:
        if by_category:
            results.sort(key=lambda val: ("%s_%5.5f" % (val['id'].split("_")[1], val['prob'])), reverse = True)
        else:
            results.sort(key=lambda val: val['prob'], reverse = True)
    else:
        if by_category:
            results.sort(key=lambda val: ("%s_%04d" % (val['id'].split("_")[1], int(val['id'].split("_")[2]))))
        else:
            results.sort(key=lambda val: int(val['id'].split("_")[2]))

    results.sort(key=lambda val: int(val['id'].split("_")[0]))

    users = 0
    nDCG_sum = 0

    for u, user_rows in itertools.groupby(results, key = lambda val: val['id'].rsplit("_", 1 if by_category else 2)[0]):
        pos_count = 0
        rows = []
        row_count = 0
        iDGC = 0.0
        dgc = 0.0
        r1_rows = []
        for i, row in enumerate(r1_rows):
            row_count += 1
            rows.append(row)
            if row['pos'] > 0:
                pos_count += 1
                iDGC += 1 / (math.log(float(pos_count + 1), 2))
                dgc +=  1 / (math.log(float(i + 2), 2))
            # if i == 9:
            #     break

        if pos_count > 0 and row_count > 5 and pos_count != row_count:
        # if pos_count > 0:
            users += 1
            nDCG_sum += (dgc / iDGC)
            if users < print_users:
                for row in rows:
                    print(row)
                print("nDGC: %6.4f\n\n" % (dgc / iDGC))

    print("%d users. Avg nGDC: %5.3f" % (users, nDCG_sum / users))


def compute_groups(dataset):
    groups = []
    count = 0
    last = dataset._ids[0].split("_")[0]
    for label in dataset._ids:
        group_id = label.split("_")[0]
        if last == group_id:
            count += 1
        else:
            groups.append(count)
            last = group_id
            count = 1
    groups.append(count)

    return groups

def explode_results(results_grouped):
    output_results = []
    for group in results_grouped:
        for entry in group:
            output_results.append(entry)

    return output_results

def compute_NGDC_ranked(results, for_model = True, print_users = 0):
    users = 0
    nDCG_sum = 0
    for group in results:
        pos_count = 0
        iDGC = 0
        dgc = 0

        r1_rows = []
        for row in group:
            if "R1" in row['id']:
                r1_rows.append(row)
        group = r1_rows

        if for_model:
            group.sort(key=lambda val: val['prob'], reverse = True)
        else:
            group.sort(key=lambda val: int(val['id'].split("_")[2]))
        for i, row in enumerate(group):
            if row['pos'] > 0.0:
                iDGC += (2**row['pos'] - 1) / (math.log(float(pos_count + 2), 2))
                dgc +=  (2**row['pos'] - 1) / (math.log(float(i + 2), 2))
                pos_count += 1
                # if i == 9:
                #     break

        if pos_count > 0:
            users += 1
            nDCG_sum += (dgc / iDGC)
            if users < print_users:
                if for_model:
                    group_al = sorted(group, key=lambda val: int(val['id'].split("_")[2]))
                    for row, row_al in zip(group, group_al):
                        print("%5.3f %2d | %2d" % (row['prob'], row['pos'], row_al['pos']))

                print("nDGC: %6.4f\n\n" % (dgc / iDGC))

    print("%d users. Avg nGDC: %5.3f" % (users, nDCG_sum / users))


def compute_likes_at(results, for_model = True, by_category = False, stop_at_index = 20, print_users = 0):
    if for_model:
        if by_category:
            results.sort(key=lambda val: ("%s_%5.5f" % (val['id'].split("_")[1], val['prob'])), reverse = True)
        else:
            results.sort(key=lambda val: val['prob'], reverse = True)
    else:
        if by_category:
            results.sort(key=lambda val: ("%s_%04d" % (val['id'].split("_")[1], int(val['id'].split("_")[2]))))
        else:
            results.sort(key=lambda val: int(val['id'].split("_")[2]))

    results.sort(key=lambda val: int(val['id'].split("_")[0]))

    users = 0
    pos_sum = 0

    for u, user_rows in itertools.groupby(results, key = lambda val: val['id'].rsplit("_", 1 if by_category else 2)[0]):
        has_non_E = False
        pos_count = 0
        rows = []
        row_count = 0
        for i, row in enumerate(user_rows):
            if "E" not in row['id'].split("_")[1]:
                has_non_E = True
            row_count +=1
            rows.append(row)
            if i < stop_at_index:
                if row['pos'] == 1.0:
                    pos_count += 1

            if i >= stop_at_index and (by_category or has_non_E):
                users += 1
                pos_sum += pos_count
                break

    if pos_count > 0 and row_count > 5 and pos_count != row_count:
            if users < print_users:
                for row in rows:
                    print(row)

    print("%d users. Avg pos_sum: %5.3f" % (users, float(pos_sum) / users))

"""
Creates a RowHandler for parsing AL data and handing to XGBoost.
"""
def create_row_handler(features, filter, label_extractor, load_from=None):
    # Set this to experiment with weighting.
    weight_extractor = None
    row_handler = RowHandler(
        label_bucket=ValueBucket("positive_interest", label_extractor, has_null=False),
        id_extractor=SearchIdExtractor(),
        weight_extractor=weight_extractor,
        load_from=load_from,
        extra_fields=[])
    if load_from is None:
        # print "Excluded features: %s" % str(EXCLUDED_FEATURES)
        buckets = filter.create_buckets(features)
        for bucket in buckets:
            row_handler.add_bucket(bucket)

    return row_handler

def train_and_push(feature_json, model_name, feature_filter, group_function = None, 
    label_extractor = InterestStateExtractor(),  **train_kwrd):
    rh, datasets = get_dataset(feature_json, feature_filter, label_extractor, run_analysis=False, rows=None)
    classifier = xgb.XgbClassifier(group_function = group_function)
    classifier.train(datasets, **train_kwrd)
    results = classifier.get_sorted_results(datasets.test)
    if group_function is not None:
        results = explode_results(results)
    spec = model_utils.XgbModelSpec(model_name, ".")
    spec.save(classifier, datasets._row_handler, results)
    model_utils.save_results_as_csv(model_name + '.csv', 'id', 'pred', results)
    spec.upload(os.path.join(model_utils.GCS_BUCKET, model_name))



import argparse
import boto3
import json

# AWS APIs only accept cannonical IDs for ACLS, not ARNS!
canonical_id = "f67caabe62ff981b4c0e480246704591414566fce3fc8f91a5cab49bfd4ee688"

new_grant = {
    "Grantee" : {
        "Type": "CanonicalUser",
        "ID": canonical_id,        
    },
    "Permission": "READ"
}

def grant_acl(bucket, key, s3):
    acl = s3.get_object_acl(Bucket = bucket, Key = key)
    del acl["ResponseMetadata"]
    # print json.dumps(acl, indent = 4)
    print("Updating: %s" % key)
    acl["Grants"].append(new_grant)
    s3.put_object_acl(Bucket = bucket, Key = key, AccessControlPolicy = acl)


parser = argparse.ArgumentParser(description='Grants S3 access to s3_data_user')
parser.add_argument('--prefix', type = str, help = 'Prefix of S3 bucket', default = "")
# Assumes AWS S3 credentials are available in local env.
# Uncomment here and below if that's not the case
# parser.add_argument('--s3_key_id', type=str, help='AWS user id for S3', required=True)
# parser.add_argument('--s3_secret_key', type=str, help='AWS secret key for S3', required=True)
parser.add_argument('--s3_bucket', type = str, help = 'S3 bucket to use', required = True)
parser.add_argument('--start_after', type = str, help = 'Start Key (useful if update fails in middle)', default = "")
args = parser.parse_args()

session = boto3.Session(
    # aws_access_key_id=args.s3_key_id,
    # aws_secret_access_key=args.s3_secret_key,
)
s3 = session.client('s3')


# First grant access to bucket
bucket_acl = s3.get_bucket_acl(Bucket = args.s3_bucket)
del bucket_acl["ResponseMetadata"]
bucket_acl["Grants"].append(new_grant)
s3.put_bucket_acl(Bucket = args.s3_bucket, AccessControlPolicy = bucket_acl)

# Iterate over items until there are no more items
current_key = args.start_after
while True:
    response = s3.list_objects_v2(Bucket=args.s3_bucket, Prefix=args.prefix, StartAfter = current_key)
    if "Contents" not in response:
        break
    contents = response["Contents"]
    for entry in contents:
        grant_acl(args.s3_bucket, entry["Key"], s3)

    current_key = contents[-1]["Key"]



import argparse
import bytegain.data_import.segment_utils as segment_utils
import bytegain.custom.data_import.db_utils as db_utils
import bytegain.custom.data_import.schema_info as schema_info_module
from collections import defaultdict

def parse_where(arg):
    return arg.split(":")

parser = argparse.ArgumentParser(description='Prints data on segment schema')
parser.add_argument('--src_url', type=str, help='segment db url', required=True)
parser.add_argument('--src_schema', type = str, help = 'segment schema', required = True)
parser.add_argument('--start_date', type = segment_utils.valid_date, help = 'Start date of data to copy', default = None)
parser.add_argument('--end_date', type = segment_utils.valid_date, help = 'End date of data to copy', default = None)
parser.add_argument('--global_where', type = str, help = "Global where clause")
parser.add_argument('--where_dict', type=parse_where, nargs="+", help = "Comma separated list of table:WHERE clauses")
parser.add_argument('--out', type=str, required = True, help = "Output file for data")
parser.add_argument('--no_groups', action='store_true', help = "If set then no grouping is done")

args = parser.parse_args()

where_dict = defaultdict(lambda: None, [(x[0], x[1]) for x in args.where_dict])
where = db_utils.create_date_clause(args.start_date, args.end_date)
if len(where) > 0:
    where = "WHERE %s" % where

src_conn_factory = db_utils.ConnectionFactory(args.src_url)
tracks = schema_info_module.TableInfo(args.src_schema, "tracks", db_utils.and_where(args.global_where, where_dict["tracks"]))
tracks = db_utils.compute_table_info(src_conn_factory, tracks, where, no_groups=args.no_groups)

pages = schema_info_module.TableInfo(args.src_schema, "pages", db_utils.and_where(args.global_where, where_dict["pages"]))
pages = db_utils.compute_table_info(src_conn_factory, pages, where, no_groups=args.no_groups)

tables = db_utils.get_segment_tables(src_conn_factory, args.src_schema)
table_infos = {}
skipped_tables = []
for table in tables:
    if table == "pages":
        continue

    table_where = db_utils.and_where(args.global_where, where_dict[table])

    table_info = schema_info_module.TableInfo(args.src_schema, table, table_where)
    info = db_utils.compute_table_info(src_conn_factory, table_info, where, exclude_common = tracks, no_groups=args.no_groups)
    if not info:
        skipped_tables.append(table)
        continue
    table_infos[table] = info
    # info.print_pretty()

if skipped_tables:
    print("Skipped tables: %s" % ' '.join(skipped_tables))

schema_info = schema_info_module.SchemaInfo(args.src_schema, tracks, pages, table_infos)
schema_info_module.save_schema(args.out, schema_info, compact=True)



"""
Plots AUC for binary classifier
"""

import numpy
import numpy.ma
import matplotlib
import sys
import csv
from scipy.stats import binom
from sklearn import metrics
from enum import Enum

matplotlib.use('Agg')

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker


def create_total_pos(results):
    """Returns a list of cumulative fraction of positive examples.
    The lengt of the list is max(1, number of negative examples)"""
    length = len(results)
    #total_pos = numpy.zeros([length], dtype=numpy.float)
    total_pos = [0.0]
    for i in range(0, length):
        pos = results[i]['label']
        if pos:
            total_pos[-1] += 1
        else:
            total_pos.append(total_pos[-1])
    pos_count = total_pos[-1]
    if length == 0 or pos_count == 0:
        raise Exception("WARNING: bad data for results,total_pos: %d %d"% (length, pos_count))

    total_pos = numpy.array(total_pos)
    total_pos /= pos_count
    return total_pos

def compute_auc(results):
    y_true = [result['label'] for result in results]
    y_pred = [result['probability'] for result in results]
    # y_pred_label = [1 if result['probability'] > 0.5 else 0 for result in results]
    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_pred)
    return metrics.auc(fpr, tpr)


def compute_metrics(results, threshold=0.5):
    y_true = [result['label'] for result in results]
    y_pred = [result['probability'] for result in results]
    y_pred_label = [1 if result['probability'] > threshold else 0 for result in results]

    metrs = {
        'precision': metrics.precision_score(y_true, y_pred_label),
        'recall': metrics.recall_score(y_true, y_pred_label),
        'accuracy': metrics.accuracy_score(y_true, y_pred_label)
    }

    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_pred)
    metrs['auc'] = metrics.auc(fpr, tpr)
    # Find fpr and tpr for a given threshold
    thresholds = thresholds.tolist()
    index = binary_search_reverse(thresholds, threshold)
    metrs['false_positive_rate'], metrs['true_positive_rate'] = fpr[index], tpr[index]

    # All metrics are numpy floats - convert them to python float
    for key, value in list(metrs.items()):
        metrs[key] = value.item()

    conf_mtr = metrics.confusion_matrix(y_true, y_pred_label)
    metrs['confusion_mtr'] = numpy.ndarray.tolist(conf_mtr)
    return metrs


# Binary search in a reversely sorted array
def binary_search_reverse(a, x):
    lo = 0
    hi = len(a)
    while lo < hi:
        mid = (lo+hi)//2
        mid_element = a[mid]
        if mid_element > x:
            lo = mid+1
        else:
            hi = mid
    return lo

def sort_results(results):
    results.sort(key=lambda v: v['probability'], reverse=True)


"""
Computes actual & predicted values for each "bucket" in the sorted
result list.
bucket_fractions is an array of floats, where each value represents
the fraction of results to use for that bucket.
bucket_fractions should sum to <1 (the remainder is used for the last bucket)
"""


def get_buckets(results, bucket_fractions):
    count = len(results)
    # Last bucket is "remainer"
    bucket_count = len(bucket_fractions) + 1
    breakpoints = [0] * bucket_count
    fraction_sum = 0.0
    for i in range(len(bucket_fractions)):
        fraction_sum += bucket_fractions[i]
        breakpoints[i] = int(fraction_sum * count)

    breakpoints[len(bucket_fractions)] = count
    actual = []
    predicted = []
    confidence = []
    prev_end = 0
    for i in range(bucket_count):
        bucket_size = breakpoints[i] - prev_end
        pred_sum = 0.0
        actual_sum = 0.0
        for j in range(prev_end, breakpoints[i]):
            pred_sum += results[j]['probability']
            if results[j]['label']:
                actual_sum += 1.0

        actual_val = actual_sum / bucket_size
        actual.append(actual_val)
        predicted.append(pred_sum / bucket_size)
        conf = binom.interval(0.95, bucket_size, actual_val)
        confidence.append(conf[1] / bucket_size - actual_val)
        prev_end = breakpoints[i]

    return actual, predicted, confidence


def plot_predicted_vs_actual(filename, results, smoothing=500):
    print("Plot datapoints: %d Smoothing: %d" % (len(results), smoothing))
    length = len(results)
    preds = []
    leases = []
    np_lease = numpy.zeros([length])
    np_prob = numpy.zeros([length])
    for i in range(length):
        np_prob[i] = results[i]['probability']
        if results[i]['label']:
            np_lease[i] = 1.0

    for i in range(length - smoothing):
        leases.append(numpy.ma.average(np_lease[i:i + smoothing]))
        preds.append(numpy.ma.average(np_prob[i:i + smoothing]))
    do_plot_predicted_vs_actual(filename, preds, leases)


def do_plot_predicted_vs_actual(filename, predictions, actual, x_values, use_percent=True, x_axis_label=None, y_axis_label=None):
    length = len(predictions)
    assert length > 1
    # print "input length: %d" % length
    plt.clf()
    if use_percent:
        predictions = [x * 100 for x in predictions]
        actual = [x * 100 for x in actual]

    max_val = max([predictions[0], actual[0], predictions[-1], actual[-1]]) * 1.1
    plt.subplots_adjust(left=0.1, right=0.97, top=0.98, bottom=0.09, wspace=0.0, hspace=0.0)
    # print ("length: %s" % str(length + (length / 10 - 1)))
    xrange = numpy.arange(0, 11)
    xrange = xrange * (length) / 10.0
    # print "predictions: %s" % str(predictions)
    plt.xticks(xrange)
    ax = plt.gca()
    if use_percent:
        ax.xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, p: "%3d%%" % (float(x) * 100 / (length))))
    else:
        ax.xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, p: "%3.1f" % (float(x) / (length))))
    ax.set_ylim([0, max_val])
    ax.set_autoscale_on(False)
    plt.grid(b=True, which='major', axis='y', color='0.3', linestyle='-', alpha=0.3, linewidth=0.3)
    lines = []
    if x_axis_label != None:
        plt.xlabel(x_axis_label)

    if y_axis_label != None:
        plt.ylabel(y_axis_label)
    # Reverse x values so plot goes Up & to the Right.
    x_values = [length - x for x in x_values]
    line, = plt.plot(x_values, predictions, "b-", linewidth=1, antialiased=True, label="predictions")
    lines.append(line)
    if actual != None:
        # plot_x = numpy.arange(0, length, length / float(len(actual)))
        line, = plt.plot(x_values, actual, "g-", linewidth=1, antialiased=True, label="actual")
        lines.append(line)
    plt.legend(handles=lines)
    plt.savefig(filename, dpi=300)


def get_predicted_vs_actual(results, points=None, min_positive_per_point = 100):
    length = len(results)

    # Compute the minimum # of samples per point.
    # We want 1000 positive (or negative if they are rarer) examples per point.
    actual = 0
    for j in range(0, length):
        if results[j]['label']:
            actual += 1.0

    if actual < min_positive_per_point * 3:
        min_positive_per_point = actual / 3

    if points is not None:
        delta = float(length)/points
        stop_points = [int(i * delta) - 1 for i in range(1, points+1)]
    else:
        stop_points = []

    min_distance = length / 100
    actual_points = []
    pred_points = []
    ece = 0.0  # Expected calibration error
    x_values = []
    last_point = 0
    pred = 0.0
    actual = 0.0
    # Compute positive rate in top 10% bucket
    actual_ten_percent = 0.0
    count_ten_percent = int(len(results)*0.1)
    for i, point in enumerate(results):
        pred += results[i]['probability']
        if results[i]['label']:
            if i < count_ten_percent:
                actual_ten_percent += 1.0
            actual += 1.0

        if (((actual > min_positive_per_point and i - last_point > min_distance) or i == len(results) - 1) and points is None) or i in stop_points:
            count = i - last_point + 1
            accuracy = actual / count  # accuracy
            confidence = pred / count  # confidence
            ece += count * abs(accuracy - confidence)

            # Hack so end points are 0%/100%
            x_values.append( float(i) if i == len(results) -1 else float(last_point))
            actual_points.append(accuracy)
            pred_points.append(confidence)
            last_point = i
            actual = 0.0
            pred = 0.0

    ece = ece / length

    print("Positive rate in top 10 percent: %s " % (actual_ten_percent/count_ten_percent))

    x_values = [(x * len(x_values)) / length for x in x_values]
    # print "Expected Calibration Error: %.4f." % ece
    return ece, pred_points, actual_points, x_values


def plot_predicted_vs_actual2(filename, results, points=None, use_percent=True, x_axis_label=None, y_axis_label=None):
    ece, pred_points, actual_points, x_values = get_predicted_vs_actual(results, points)
    do_plot_predicted_vs_actual(filename, pred_points, actual_points, x_values, use_percent, x_axis_label, y_axis_label)
    return ece


class PlotType(Enum):
    percent = 'percent'
    odds_ratio = 'odds_ratio'
    percent_of_total = 'percent_of_total'

    def __str__(self):
        return self.value

def plot_by_segment(filename, results, y_axis_label = None, title = None, plot_type = PlotType.percent_of_total,
                    buckets = None):
    """

    :param filename: Image file to create
    :param results: Sorted result set
    :param y_axis_label: Graph label for y Axis
    :param title: Plot title
    :return: sums: Ratio/percentage of positive events in each bucket
    """
    if buckets == None:
        buckets = (0.2, 0.2, 0.2, 0.2, 0.2)
    plt.clf()
    fig, ax = plt.subplots()
    labels = ('Very High', 'High', 'Med', 'Low', 'Lowest')
    labels = labels[0:len(buckets)]
    labels_perc = []
    for i in range(len(buckets)):
        labels_perc.append(labels[i]+'\n'+str(int(buckets[i]*100))+'%')
    buckets = numpy.array(buckets)
    cumsum = numpy.cumsum(buckets)
    length = len(results)
    y_pos = numpy.arange(len(labels))
    sums = []
    current_count = 0.0
    current_sum = 0.0
    target =  cumsum[0] * length
    total_positive = 0.0
    for example in results:
        if example['label']:
            total_positive +=1

    if total_positive == 0.0:
        raise Exception("No positive results")

    print("Plot type: %s           " % plot_type)
    if plot_type == PlotType.odds_ratio:
        total = 0
        for result in results:
            if result['label']:
                total += 1
        average = total / float(length)
        multiplier = 1.0 / average
    elif plot_type == PlotType.percent:
        multiplier = 100
    elif plot_type == PlotType.percent_of_total:
        multiplier = 100
    else:
        raise Exception("Unknown plot_type: %s %s" % (plot_type, type(plot_type)))

    for i in range(length):
        current_count += 1
        if results[i]['label']:
            current_sum += 1
        if i <= target and (i+1) > target or i == length - 1:
            if plot_type == PlotType.percent_of_total:
                sums.append(current_sum / total_positive * multiplier)
            else:
                sums.append(current_sum / current_count * multiplier)

            current_sum = 0.0
            current_count = 0.0
            if len(sums) < len(buckets):
                target = cumsum[len(sums)] * length

    plt.subplots_adjust(left=0.12, right=0.98, top=0.95, bottom=0.13, wspace=0.0, hspace=0.0)
    bars = plt.bar(y_pos, sums, align='center', alpha=0.5)
    for bar in bars:
        height = bar.get_height()
        bottom, top = ax.get_ylim()
        yrange = (top - bottom)
        if height > bottom + yrange * 0.8:
            print_height = height - yrange * 0.06
        else:
            print_height = height + yrange * 0.03
        ax.text(bar.get_x() + bar.get_width()/2., print_height,
                ('%.1f%%' % height) if plot_type in (PlotType.percent_of_total, PlotType.percent) else ("%4.2fx" % height),
                ha='center', va='bottom')

    plt.xticks(y_pos, labels_perc)
    plt.xlabel("User Segment (Best to Worst)")
    # plt.ylabel(y_axis_label)
    if plot_type == PlotType.odds_ratio:
        plt.ylabel("Purchase odds vs average user")
    elif plot_type == PlotType.percent_of_total:
        plt.ylabel("Percent of total positive samples")
    else:
        plt.ylabel("Percent positive in segment")

    if title:
        plt.title(title)
    plt.savefig(filename, dpi=300)
    return sums


def plot_auc(filename, results, results2=None, results3=None, x_label=None, y_label=None,
             labels=None):
    plt.clf()
    total_pos = create_total_pos(results)
    length = len(total_pos)
    if results2 != None:
        total_pos2 = create_total_pos(results2)
    if results3 != None:
        total_pos3 = create_total_pos(results3)

    if x_label == None and y_label == None:
        plt.subplots_adjust(left=0.05, right=0.98, top=0.97, bottom=0.08, wspace=0.0, hspace=0.0)
    else:
        plt.subplots_adjust(left=0.08, right=0.98, top=0.97, bottom=0.09, wspace=0.0, hspace=0.0)

    xticks = numpy.arange(length + 0.001, step=length / 10.0)
    plt.xticks(xticks)
    plt.yticks(numpy.arange(0, 1, step=0.1))
    ax = plt.gca()
    ax.set_autoscale_on(False)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(lambda x, p: "%3.1f" % (float(x) / length)))
    plt.grid(b=True, which='major', axis='y', color='0.3', linestyle='-', alpha=0.3, linewidth=0.3)
    plt.grid(b=True, which='major', axis='x', color='0.3', linestyle='-', alpha=0.3, linewidth=0.3)
    plt.plot([0, length - 1], [0, 1], "r-", linewidth=0.2, antialiased=True)

    if y_label:
        ax.set_ylabel(y_label)

    if x_label:
        ax.set_xlabel(x_label)

    plt.plot(total_pos, "b-", linewidth=0.5, antialiased=True, label=labels[0] if labels else None)
    if results2 != None:
        plot_x = numpy.arange(0, length, length / float(len(total_pos2)))
        plt.plot(plot_x, total_pos2, "g-", linewidth=0.5, antialiased=True, label=labels[1] if labels else None)
    if results3 != None:
        plot_x = numpy.arange(0, length, length / float(len(total_pos3)))
        plt.plot(plot_x, total_pos3, "k-", linewidth=0.5, antialiased=True, label=labels[2] if labels else None)

    if labels:
        plt.legend(loc='upper left')
    plt.savefig(filename, dpi=300)



""" Holds a model configuration.

To build a ModelConfig from a SchemaInfo:

from bytegain.model_data.model_config import ModelConfig
ModelConfig.from_table_info_file('web_join2.pickle').to_json_file('model_configs/al_1.0')

"""

import pyximport

from bytegain.zap.prediction_instance import PredictionInstance, set_counter_impl

pyximport.install()

import pickle
import datetime
import json
import tensorflow as tf
import random

from bytegain.data_import.segment_utils import timestamp_to_epoch
from bytegain.zap import feature_coding
from bytegain.custom.model_data.bucket import SimpleCategoricalBucket, ValueBucket
from collections import OrderedDict
import importlib

def get_epoch_times_for_rows(rows):
    row_dt = []
    for i in range(0, len(rows)):
        row_dt.append(datetime.datetime.utcfromtimestamp(timestamp_to_epoch(rows[i]['timestamp'])))
    return row_dt


class ModelConfigFeature(object):
    # TODO(chris): Support non-categorical.
    def __init__(self, field, categories):
        self._field = field
        self._categories = categories

    @staticmethod
    def from_column_info(column_info, table_name = None):
        """ Construct a feature from ColumnInfo.

        If table_name is set, it is used to create a new feature name.

        Returns None if the column is not supported.
        """
        name = column_info.name

        # TODO(chris): Stop peeking into internal _groups.
        if not column_info.is_groupable() or not column_info._groups or column_info.required() or len(column_info._groups) < 2:
            return None
        # name = ("%s_%s" % (table_name, column_info.name)) if table_name is not None else column_info.name
        # print ("Including: %s.%s" % (table_name, column_info))
        if column_info.name == "event":
            # Hack so we capture all events since they are usually used in prediction instance
            group_limit = 0.00001
        else:
           # The minimum ratio for a group is g[1]/20 - this prevents tons of outliers on the tail.
            # It's g[1] not g[0] since g[0] is often null
            group_limit = max(column_info._groups[1][1] / 20.0, 0.0001)

        groups = [g[0] for g in column_info._groups if g[1] > group_limit]

        # If null_fraction > last_group fraction then add a None
        if column_info.null_fraction > column_info._groups[-1][1] and None not in groups:
            groups.append(None)

        if len(groups) < 2 and "char" in column_info.datatype:
            return None
        return ModelConfigFeature(name, groups)

    @staticmethod
    def merge_features(feature1, feature2):
        assert feature1._field == feature2._field
        categories = set(feature1._categories)
        categories.update(feature2._categories)
        return ModelConfigFeature(feature1._field, [g for g in categories])

    @staticmethod
    def from_dict(rep):
        return ModelConfigFeature(rep['field'], rep['categories'])

    def to_dict(self):
        return OrderedDict([
                ('field', self._field),
                ('categories', self._categories)])

    @property
    def field(self):
        return self._field

    @property
    def categories(self):
        return self._categories

class ModelConfigEventSelector(object):
    def __init__(self, fields, expression):
        self._fields = fields
        self._expression = expression

    @staticmethod
    def from_dict(rep):
        return ModelConfigEventSelector(rep['fields'], rep['expression'])

    def to_dict(self):
        return OrderedDict([
                ('fields', self._fields),
                ('expression', self._expression)])

    @property
    def fields(self):
        return self._fields

    @property
    def expression(self):
        return self._expression

class ModelConfig(object):
    """ Configuration for training a clickstreaam model from a sequence of events. """
    def __init__(self, id_field, instance_generator, blackout_seconds, negative_fraction, features, timestamp_field = 'received_at', max_events = 40):
        self._id_field = id_field
        self._instance_generator = instance_generator
        self._blackout_seconds = blackout_seconds
        self._negative_fraction = negative_fraction
        self._timestamp_field = timestamp_field
        self._features = features
        self._max_events = max_events

        # Maximum time between duplicate events
        self._dup_max_time = 3

    @staticmethod
    def from_schema_info_file(path):
        instance_generator = 'UNCONFIGURED'
        from bytegain.custom.data_import.schema_info import read_schema_from_file as read_schema_info_from_file
        schema_info = read_schema_info_from_file(path)

        # Maps from column.name -> Feature
        # Used to merge multiple columns with same name
        feature_dict = {}
        for column in schema_info.main_table.columns:
            feature = ModelConfigFeature.from_column_info(column)
            if feature:
                feature_dict[column.name] = feature
        for column in schema_info.page_table.columns:
            feature = ModelConfigFeature.from_column_info(column, schema_info.page_table.table)
            if feature:
                if column.name in feature_dict:
                    feature = ModelConfigFeature.merge_features(feature_dict[column.name], feature)
                feature_dict[column.name] = feature

        for table_info in list(schema_info.join_tables.values()):
            for column in table_info.columns:
                feature = ModelConfigFeature.from_column_info(column, table_info.table)
                if feature:
                    if column.name in feature_dict:
                        feature = ModelConfigFeature.merge_features(feature_dict[column.name], feature)
                    feature_dict[column.name] = feature

        features = sorted(list(feature_dict.values()), key = lambda v : v._field)
        return ModelConfig('UNCONFIGURED', instance_generator, 1.0, 1.0, features)

    @staticmethod
    def from_table_info_file(path):
        table_info = pickle.load(open(path, 'rb'))
        features = []
        for column in table_info.columns:
            feature = ModelConfigFeature.from_column_info(column)
            if feature:
                features.append(feature)
        return ModelConfig('UNCONFIGURED', 'UNCONFIGURED', 1.0, 1.0, 'UNCONFIGURED', features)

    @staticmethod
    def from_json_file(path):
        with open(path) as f:
            jrep = json.load(f)
            return ModelConfig.from_dict(jrep)

    @staticmethod
    def from_dict(jrep):
        features = [ModelConfigFeature.from_dict(feature_rep) for feature_rep in jrep['features']]
        return ModelConfig(jrep['idField'], jrep['instanceGenerator'],
                        jrep['blackoutSeconds'], jrep['negativeFraction'], features, jrep['timestampField'], jrep['maxEvents'])

    def set_params(self, id_field, timestamp_field, instance_generator, max_events, blackout_seconds):
        self._id_field = id_field
        self._timestamp_field = timestamp_field
        self._instance_generator = instance_generator
        self._max_events = max_events
        self._blackout_seconds = blackout_seconds

    def to_json_file(self, path):
        with open(path, 'w') as f:
            json.dump(self.to_dict(), f, indent=4)

    def to_dict(self):
        return OrderedDict([
                ('idField', self._id_field),
                ('instanceGenerator', self._instance_generator),
                ('blackoutSeconds', self._blackout_seconds),
                ('negativeFraction', self._negative_fraction),
                ('timestampField', self._timestamp_field),
                ('maxEvents', self._max_events),
                ('features', [feature.to_dict() for feature in self._features])])

    @property
    def id_field(self):
        return self._id_field

    @property
    def instance_generator(self):
        return self._instance_generator

    @property
    def blackout_seconds(self):
        """ Number of seconds before a prediction event to drop from training. """
        return self._blackout_seconds

    @property
    def negative_fraction(self):
        """ Fraction of 0-label instances to include in training. """
        return self._negative_fraction

    @property
    def timestamp_field(self):
        return self._timestamp_field

    @property
    def features(self):
        return self._features

    def get_feature_width(self):
        # Add one for "non of the above"
        return sum([len(x._categories) + 1 for x in self._features])

    def __str__(self):
        return json.dumps(self.to_dict(), indent=4)

class DictionaryRowExtractor(object):
    def __init__(self, field_name):
        self._field_name = field_name
        self._splits = field_name.split(".")
     
    def _extract(self, row, index):        
        if index == len(self._splits):
            return row
        if self._splits[index] in row:
            value = row[self._splits[index]]
            if type(value) in (tuple, list):
                results = []
                for entry in value:
                    result = self._extract(entry, index + 1)
                    if result is not None:
                        if type(result) is list:
                            results.extend(result)
                        else:
                            results.append(result)
                return results if len(results) > 0 else None
            else:
                return self._extract(value, index + 1)

    def extract(self, row):
        if len(self._splits) == 1:
            return row[self._field_name]
        else:
           return self._extract(row, 0)

    def get_fields(self):
        return [self._field_name]

class ModelConfigRowHandler(object):
    def __init__(self, model_config, csv_input=False, counter = None):
        self._counter = counter

        self._model_config = model_config
        generator_package_name, generator_name = model_config.instance_generator.rsplit(".", 1)
        package = importlib.import_module(generator_package_name)
        self._instance_generator = getattr(package, generator_name)()
        self._csv_input = csv_input
        self._buckets = []
        self._id_extractor = DictionaryRowExtractor(model_config.id_field)
        for feature in model_config.features:
            # TODO(chris): Support non-categorical
            if feature.categories:
                self._buckets.append(SimpleCategoricalBucket(feature.field, DictionaryRowExtractor(feature.field), feature.categories))
            else:
                self._buckets.append(ValueBucket(feature.field, DictionaryRowExtractor(feature.field), has_null = True))


    def get_fields(self):
        fields = []
        for bucket in self._buckets:
            fields.extend(bucket._extractor.get_fields())
        fields.append(self._model_config.timestamp_field)
        # Label and id will get added elsewhere.
        return fields

    def replace_field_names_with_index(self, cursor):
        # Instead of using a DictCursor, use a regular cursor,
        # and rewrite the extractor field names with array
        # offsets.  This speeds up data loading by about 20%.
        header = [x.name for x in cursor.description]
        self.replace_field_names_with_index_header(header)

    def replace_field_names_with_index_header(self, header):
        for bucket in self._buckets:
            bucket._extractor.update_fields_by_id(header)

    def get_id_from_row(self, row):
        return row[self._model_config.id_field]

    def add_sequence_examples(self, row_iterator, writer):
        # Restore this each time since global state is not saved in Beam
        set_counter_impl(self._counter)

        dup_count = 0
        output_prediction_instances = []
        # Make sure we have a list, and not an iterator.
        rows = list(row_iterator)
        # Sort rows by timestamp.  We skipped doing this in SQL-land to make time to first by fast.
        rows = sorted(rows, key=lambda row: row[self._model_config.timestamp_field])
        new_rows = []
        bucket_data = []
        dup_count = 0
        # Remove duplicates
        for row in rows:
            duplicate = True
            data = self.make_feature_dictionary(row)
            data['timestamp'] = timestamp_to_epoch(row[self._model_config.timestamp_field])
            if len(bucket_data) == 0:
                duplicate = False
            else:
                last_data = bucket_data[-1]
                if (data['timestamp'] - last_data['timestamp']) > self._model_config._dup_max_time:
                    duplicate = False
                else:
                    for k in data:
                        if k != 'timestamp' and data[k] != last_data[k]:
                            duplicate = False
                            break
            if not duplicate:                    
                bucket_data.append(data)
                new_rows.append(row)
            else:
                dup_count += 1
        
        rows = new_rows
        # Make downsampling deterministic.
        rng = random.Random()
        rng.seed(self.get_id_from_row(rows[0]))
        prediction_instances = self._instance_generator.get_instances(rows)

        # Don't give a user more than 100 training instances.  Just use the last 100.
        # This number is more tuned for minimizing mapreduce stragglers than anything else.
        # With the current dataset this affects about 0.35% (.0035) of users.
        # TODO(chris): Do we still need this?
        if len(prediction_instances) > 100:
            prediction_instances = prediction_instances[-100:]

        # Add a training instance for each prediction event, including all events up to that event.
        instance_count = 0
        positive_instance_count = 0
                  
        for instance_i, instance in enumerate(prediction_instances):
            # Downsample negative examples if needed.
            if instance.label == False and rng.random() >= self._model_config.negative_fraction:
                continue

            end_index = instance.end - 1
            # Get the context_page_path from the prediction event.
            prediction_context_page_path = rows[end_index]['context_page_path']

            # Limit data to the prediction event minus the blackout period.
            last_usable_event_index = end_index
            while last_usable_event_index >= 0 and self._elapsed(rows[last_usable_event_index], rows[end_index]) < self._model_config.blackout_seconds:
                last_usable_event_index -= 1
                if last_usable_event_index < 0:
                    return 0, 0, output_prediction_instances, rows

            # Create the example

            entry = self.make_sequence_example(
                rows[instance.begin:(last_usable_event_index + 1)], bucket_data[instance.begin:(last_usable_event_index + 1)],
                instance, prediction_context_page_path, '%s_%d' % (self.get_id_from_row(rows[0]), end_index))
            output_prediction_instances.append(PredictionInstance(
                instance.begin, last_usable_event_index + 1, instance.label, final_dwell = None, future_data = instance.future_data))

            writer.write(entry.SerializeToString())
            instance_count += 1
            positive_instance_count += (instance.label == True)

        # TODO(jjs) - remove everything but output_prediction_instances
        return instance_count, positive_instance_count, output_prediction_instances, rows

    def make_feature_dictionary(self, row):
        output = {}
        for bucket in self._buckets:
            value = bucket.bucketize(row)
            output[bucket._name] = value

        return output
        
    def make_sequence_example(self, rows, bucket_data, instance, prediction_context_page_path, instance_id):
        context = tf.train.Features(feature={
            'future_data' : tf.train.Feature(
                bytes_list=tf.train.BytesList(value=[str(instance.future_data if instance.future_data else "")])),
            'id': tf.train.Feature(bytes_list=tf.train.BytesList(value=[instance_id.encode('utf-8')])),
            'label': ModelConfigRowHandler._int64_feature(instance.label if instance.label is not None else 99)})


        feature_values = {'categorical': []}
        width = self._model_config.get_feature_width()
        for row, row_bucket_data in zip(rows, bucket_data):
            offset = 0
            indicies = [0] * width
            # Join all feature buckets into 1 feature which has all features.
            # The output is basically the set of all indicies in the output vector that should
            # be turned on.

            for bucket in self._buckets:
                value = row_bucket_data[bucket._name]
                for entry in value:
                    # Add one to account for -1 "NONE" value
                    indicies[offset + entry + 1] = 1

                offset += len(bucket._category_value_map) + 1

            feature_values['categorical'].append(tf.train.Feature(int64_list=tf.train.Int64List(value=indicies)))

        # Add time before next event.
        # row_epochs = [timestamp_to_epoch(row[self._model_config.timestamp_field]) for row in rows]
        # TODO(chris): Move this out into a feature type in ModelConfig?
        feature_values['dwell'] = []
        for i in range(len(rows) - 1):
            delta = bucket_data[i + 1]['timestamp'] - bucket_data[i]['timestamp']
            feature_values['dwell'].append(feature_coding.compute_dwell_feature(delta))

        feature_values['dwell'].append(feature_coding.compute_dwell_feature(instance.final_dwell))

        # Add current_item
        feature_values['current_cpp'] = []
        for row in rows:
            context_page_path = row['context_page_path']
            if context_page_path == '':
                feature = [0, 0, 1]
            elif context_page_path == prediction_context_page_path:
                feature = [1, 0, 0]
            else:
                feature = [0, 1, 0]
            feature_values['current_cpp'].append(tf.train.Feature(int64_list=tf.train.Int64List(value=feature)))

        # Telescoping is an attenuation feature [N, T] = Dropped N events and T seconds after this event.
        # Currently: If we have more than E (20) events, keep the first one, the last E-1, and telescope the rest.
        TELESCOPE_TO_LENGTH = self._model_config._max_events
        feature_values['telescope'] = [tf.train.Feature(int64_list=tf.train.Int64List(value=[0, 0])) for i in range(min(len(rows), TELESCOPE_TO_LENGTH))]
        if len(rows) > TELESCOPE_TO_LENGTH:
            # First cut out events
            for feature in list(feature_values.keys()):
                feature_values[feature] = [feature_values[feature][0]] + feature_values[feature][-(TELESCOPE_TO_LENGTH-1):]

            # Then set the telescope feature on the first event.
            feature_values['telescope'][0] = tf.train.Feature(int64_list=tf.train.Int64List(value=[len(rows) - TELESCOPE_TO_LENGTH, bucket_data[-(TELESCOPE_TO_LENGTH-1)]['timestamp'] - bucket_data[1]['timestamp']]))

        # Example code for adding a time_to_move feature.
        # Don't forget to also uncomment the corresponding lines in dataset_pipeline.py.
        #
        # feature_values['time_to_move'] = []
        # for row in rows:
        #     if row['event'] == 'quiz_step' and row['move_in_date']:
        #         timestamp = datetime.datetime.strptime(row['timestamp'][:19], '%Y-%m-%d %H:%M:%S')
        #         move_in_date = datetime.datetime.strptime(row['move_in_date'][:19], '%Y-%m-%d %H:%M:%S')
        #         time_until_move = move_in_date - timestamp
        #         feature = [time_until_move.days, 0]
        #     else:
        #         feature = [0, 1]
        #    feature_values['time_to_move'].append(tf.train.Feature(int64_list=tf.train.Int64List(value=feature)))

        feature_lists = tf.train.FeatureLists(feature_list={k: tf.train.FeatureList(feature=v) for k, v in list(feature_values.items())})

        return tf.train.SequenceExample(context=context, feature_lists=feature_lists)

    @staticmethod
    def _int64_feature(value):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

    def _elapsed(self, first_row, second_row):
        second_timestamp = second_row[self._model_config.timestamp_field]
        first_timestamp = first_row[self._model_config.timestamp_field]
        if self._csv_input:
            first_timestamp = timestamp_to_epoch(first_timestamp)
            second_timestamp = timestamp_to_epoch(second_timestamp)
        return second_timestamp - first_timestamp



from bytegain.custom.cust.apartmentlist.user_value_model_bg_events.bq_utils import create_table
import argparse


def get_import_files(source_uri, schema, table_id):
    prefix = '%s/%s_%s/%s.%s.' % (source_uri, schema, table_id, schema, table_id)
    files = '%s*_part_00.gz' % prefix
    schema_file = '%s.pickle' % prefix
    return (files, schema_file)


if __name__=='__main__':
    parser = argparse.ArgumentParser("Creates (or overwrites) AL BQ tables from external data")
    parser.add_argument('--source_uri', required=True, help='GCS location of external files, '
                                                            'e.g. gs://bg-al-ltv-data/2018-10-01--2019-05-01/')
    args = parser.parse_args()

    project = 'bg-event-ingester'
    dataset_id = 'al'

    column_map = {'user_id': {'name': 'userId', 'type': 'STRING'},
                  'created_at': {'name': 'timestamp'}}


    def create_al_bq_table(schema, table_id):
        source_uris, schema_file = get_import_files(args.source_uri, schema, table_id)
        create_table(project=project, dataset_id=dataset_id, table_id=table_id,
                     source_uris=source_uris, schema_file_uri=schema_file, column_map=column_map)

    # Table with leases
    create_al_bq_table('etl', 'acappella_leases')
    # Sign up data
    create_al_bq_table('web', 'modals')


import numpy
import csv
import time
from bytegain.custom.model_data.dataset import OnceThrough

class LstmField(object):
    def __init__(self, width):
        self._width = width

    def is_sparse(self):
        return self._width > 0

class LstmCsvSpec(object):
    def __init__(self, fields, max_events):
        self._fields = fields
        self._max_events = max_events
        self._width = 0
        for field in self._fields:
            if field.is_sparse():
                self._width += field._width
            else:
                self._width += 1

    def max_events(self):
        return self._max_events

    def max_width(self):
        return self._width * self._max_events

    def sparse_width(self):
        return len(self._fields) * self._max_events

    def field_widths(self):
        return [x._width for x in self._fields]

    def create_batch_data(self, batch):
        batch_size = len(batch)
        output = numpy.zeros([batch_size, self._max_events, self._width], numpy.int)

        for row_index, row in enumerate(batch):
            for i in range (self._max_events):
                input_offset = i * len(self._fields)
                output_offset = 0
                for field_index, field in enumerate(self._fields):
                    value = row[input_offset + field_index]
                    if field.is_sparse():
                        output[row_index][i][value + output_offset] = 1
                        output_offset += field._width
                else:
                    output[row_index][i][output_offset] = value
                    output_offset += 1

        return output

    def parse_csv_row(self, row):
        id = int(row[0])
        leased = int(row[1])
        events = numpy.zeros(self._max_events * len(self._fields))
        offset = 0
        event_count = 0
        for event in row[2:][-self._max_events:]:
            event_count += 1
            event_fields = event.split(",")
            for bucket, field in zip(self._fields, event_fields):
                int_value = int(field)
                # We use -1 & -2 to indicate null, or unknown, but for sparse need all >=0
                if bucket.is_sparse():
                    int_value += 2
                    if int_value > bucket._width or int_value < 0:
                        raise Exception("Invalid sparse value %d. Max width: %d" % (int_value, bucket._width))
                events[offset] = int_value
                offset += 1

        return id, leased, events, event_count

AL_SPEC = LstmCsvSpec([LstmField(16), LstmField(5), LstmField(0)], 20)

class LstmDataset(object):
    def __init__(self, filename, lstm_spec, test_fraction = 0.1, random_seed=83243312):
        """
        Datset for LSTM data
        """
        start_time = time.time()
        self._spec = lstm_spec

        self._data = []
        self._ids = []
        self._labels = []
        self._event_counts = []
        with open(filename, "r") as f:
            reader = csv.reader(f, delimiter='|')
            for row in reader:
                id, label, events, event_count = lstm_spec.parse_csv_row(row)
                self._ids.append(id)
                self._labels.append(label)
                self._event_counts.append(event_count)
                self._data.append(events)

        self._ids = numpy.array(self._ids, dtype = numpy.int)
        self._labels = numpy.array(self._labels, dtype = numpy.float)
        self._data = numpy.array(self._data, dtype = numpy.int)
        self._event_counts = numpy.array(self._event_counts, dtype = numpy.int)
        self._ids, self._labels, self._data, self._event_counts = shuffle([self._ids, self._labels, self._data, self._event_counts])
        self._test_len = int(len(self._ids) * test_fraction)
        self._partition_count = 0
        numpy.random.seed(random_seed)
        self._partition()
        print("Loaded %d rows in %ds" % (len(self._ids), time.time() - start_time))

    def event_width(self):
        return self._spec._width

    def max_events(self):
        return self._spec._max_events

    def _partition(self):
        test_start = self._partition_count * self._test_len
        test_end = (self._partition_count + 1) * self._test_len
        self.test = LstmSet(self._spec, self._data[test_start : test_end], self._event_counts[test_start : test_end], self._labels[test_start : test_end], self._ids[test_start : test_end],
                self.event_width(), self.max_events())
        if self._partition_count != 0:
            data_start, counts_start, label_start, ids_start = self._data[0:test_start], self._event_counts[0:test_start], self._labels[0:test_start], self._ids[0:test_start]
            data_end, counts_end, label_end, ids_end = self._data[test_end:], self._event_counts[test_end:], self._labels[test_end:], self._ids[test_end:]

            self.train = LstmSet(self._spec, numpy.concatenate((data_start, data_end)), numpy.concatenate((counts_start, counts_end)), numpy.concatenate((label_start, label_end)), numpy.concatenate((ids_start, ids_end)),
             self.event_width(), self.max_events())
        else:
            self.train = LstmSet(self._spec, self._data[test_end:], self._event_counts[test_end:], self._labels[test_end:], self._ids[test_end:],
             self.event_width(), self.max_events())

    def repartition(self):
        self._partition_count += 1
        if (self._partition_count + 0.9) * self._test_fraction > 1.0:
            self._partition_count = 0
        self._partition()

def shuffle(arrays):
    vals = []
    perm = numpy.random.permutation(len(arrays[0]))

    for val in arrays:
        vals.append(val[perm])

    return vals

class LstmSet(object):
    def __init__(self, lstm_spec, data, event_counts, labels, ids, event_width, max_events):
        """Construct a DataSet.
        """
        self._lstm_spec = lstm_spec
        self._event_width = event_width
        self._max_events = max_events
        self._data = data
        self._event_counts = event_counts
        self._labels = labels
        self._ids = ids
        self._num_examples = len(data)
        self._epochs_completed = 0
        self._index_in_epoch = 0

    def once_through(self, batch_size, max_examples = -1):
        # if max_examples > 0:
        #       expanded = self._lstm_spec.create_batch_data(self._data[0:max_examples])
        # else:
        #       expanded = self._lstm_spec.create_batch_data(self._data)

        return OnceThrough([self._data, self._event_counts, self._labels, self._ids], batch_size, max_examples)

    @property
    def data(self):
        return self._data

    @property
    def event_counts(self):
        return self._event_counts

    @property
    def labels(self):
        return self._labels

    @property
    def ids(self):
        return self._ids

    @property
    def num_examples(self):
        return self._num_examples

    @property
    def epochs_completed(self):
        return self._epochs_completed

    def next_batch(self, batch_size):
        """Return the next `batch_size` examples from this data set."""
        start = self._index_in_epoch
        self._index_in_epoch += batch_size
        if self._index_in_epoch > self._num_examples:
            # Finished epoch
            self._epochs_completed += 1
            self._data, self._labels
            # Shuffle the data
            self._data, self._event_counts, self._labels, self._ids = shuffle([self._data, self._event_counts, self._labels, self._ids])

            # Start next epoch
            start = 0
            self._index_in_epoch = batch_size
            assert batch_size <= self._num_examples
        end = self._index_in_epoch
        # batch_data = self._lstm_spec.create_batch_data(self._data[start:end])
        return self._data[start:end], self._event_counts[start:end], self._labels[start:end]
